# test_bayes_classifier.py
# Contact: Jacob Schreiber <jmschreiber91@gmail.com>

import os
import numpy
import torch
import pytest

from pomegranate.bayes_classifier import BayesClassifier
from pomegranate.distributions import Exponential

from .distributions._utils import _test_initialization_raises_one_parameter
from .distributions._utils import _test_initialization
from .distributions._utils import _test_predictions
from .distributions._utils import _test_efd_from_summaries
from .distributions._utils import _test_raises

from .tools import assert_raises
from numpy.testing import assert_array_almost_equal


MIN_VALUE = 0
MAX_VALUE = None
VALID_VALUE = 1.2


@pytest.fixture
def X():
	return [[1, 2, 0],
	     [0, 0, 1],
	     [1, 1, 2],
	     [2, 2, 2],
	     [3, 1, 0],
	     [5, 1, 4],
	     [2, 1, 0],
	     [1, 0, 2],
	     [1, 1, 0],
	     [0, 2, 1],
	     [0, 0, 0]]


@pytest.fixture
def X_masked(X):
	mask = torch.tensor(numpy.array([
		[False, True,  True ],
		[True,  True,  False],
		[False, False, False],
		[True,  True,  True ],
		[False, True,  False],
		[True,  True,  True ],
		[False, False, False],
		[True,  False, True ],
		[True,  True,  True ],
		[True,  True,  True ],
		[True,  False, True ]]))

	X = torch.tensor(numpy.array(X))
	return torch.masked.MaskedTensor(X, mask=mask)


@pytest.fixture
def w():
	return [[1], [2], [0], [0], [5], [1], [2], [1], [1], [2], [0]]


@pytest.fixture
def y():
	return [1, 0, 0, 1, 1, 0, 0, 0, 0, 1, 1]


@pytest.fixture
def model():
	d = [Exponential([2.1, 0.3, 0.1]), Exponential([1.5, 3.1, 2.2])]
	return BayesClassifier(d, priors=[0.7, 0.3])


###


def test_initialization():
	d = [Exponential(), Exponential()]
	model = BayesClassifier(d)

	_test_initialization(model, None, "priors", 0.0, False, None)
	assert_raises(AttributeError, getattr, model, "_w_sum")
	assert_raises(AttributeError, getattr, model, "_log_priors")


def test_initialization_raises():
	d = [Exponential(), Exponential()]

	assert_raises(TypeError, BayesClassifier)
	assert_raises(ValueError, BayesClassifier, d, [0.2, 0.2, 0.6])
	assert_raises(ValueError, BayesClassifier, d, [0.2, 1.0])
	assert_raises(ValueError, BayesClassifier, d, [-0.2, 1.2])

	assert_raises(ValueError, BayesClassifier, Exponential)
	assert_raises(ValueError, BayesClassifier, d, inertia=-0.4)
	assert_raises(ValueError, BayesClassifier, d, inertia=1.2)
	assert_raises(ValueError, BayesClassifier, d, inertia=1.2, frozen="true")
	assert_raises(ValueError, BayesClassifier, d, inertia=1.2, frozen=3)
	

def test_reset_cache(X, y):
	d = [Exponential(), Exponential()]
	
	model = BayesClassifier(d)
	model.summarize(X, y)
	
	assert_array_almost_equal(model._w_sum, [6.0, 5.0])
	assert_array_almost_equal(model._log_priors, [-0.693147, -0.693147])

	model._reset_cache()
	assert_array_almost_equal(model._w_sum, [0.0, 0.0])
	assert_array_almost_equal(model._log_priors, [-0.693147, -0.693147])	


def test_initialize(X):
	d = [Exponential(), Exponential()]
	model = BayesClassifier(d)

	assert model.d is None
	assert model.k == 2
	assert model._initialized == False
	assert_raises(AttributeError, getattr, model, "_w_sum")
	assert_raises(AttributeError, getattr, model, "_log_priors")

	model._initialize(3)
	assert model._initialized == True
	assert model.priors.shape[0] == 2
	assert model.d == 3
	assert model.k == 2
	assert_array_almost_equal(model.priors, [0.5, 0.5])
	assert_array_almost_equal(model._w_sum, [0.0, 0.0])

	model._initialize(2)
	assert model._initialized == True
	assert model.priors.shape[0] == 2
	assert model.d == 2
	assert model.k == 2
	assert_array_almost_equal(model.priors, [0.5, 0.5])
	assert_array_almost_equal(model._w_sum, [0.0, 0.0])

	d = [Exponential([0.4, 2.1]), Exponential([3, 1]), Exponential([0.2, 1])]
	model = BayesClassifier(d)
	assert model._initialized == True
	assert model.d == 2
	assert model.k == 3

	model._initialize(3)
	assert model._initialized == True
	assert model.priors.shape[0] == 3
	assert model.d == 3
	assert model.k == 3
	assert_array_almost_equal(model.priors, [1./3, 1./3, 1./3])
	assert_array_almost_equal(model._w_sum, [0.0, 0.0, 0.0])



###


def test_emission_matrix(model, X):
	e = model._emission_matrix(X)

	assert_array_almost_equal(e, 
		[[ -4.7349,  -4.8411],
         [ -7.5921,  -3.9838],
         [-21.4016,  -5.4276],
         [-25.2111,  -6.4169],
         [ -2.3540,  -5.8519],
         [-43.3063,  -9.0034],
         [ -1.8778,  -5.1852],
         [-18.0682,  -5.1051],
         [ -1.4016,  -4.5185],
         [-14.2587,  -4.6290],
         [  2.4079,  -3.5293]], 4)
	assert_array_almost_equal(e[:, 0], model.distributions[0].log_probability(X) 
		- 0.3567, 4)
	assert_array_almost_equal(e[:, 1], model.distributions[1].log_probability(X) 
		- 1.2040, 4)


def test_emission_matrix_raises(model, X):
	_test_raises(model, "_emission_matrix", X, min_value=MIN_VALUE)

	d = [Exponential(), Exponential()]
	model = BayesClassifier(d)
	_test_raises(model, "_emission_matrix", X, min_value=MIN_VALUE)


def test_log_probability(model, X):
	logp = model.log_probability(X)
	assert_array_almost_equal(logp, [-4.0935, -3.9571, -5.4276, -6.4169, 
		-2.3241, -9.0034, -1.8418, -5.1051, -1.3582, -4.6289,  2.4106], 4)


def test_log_probability_raises(model, X):
	_test_raises(model, "log_probability", X, min_value=MIN_VALUE)

	d = [Exponential(), Exponential()]
	model = BayesClassifier(d)
	_test_raises(model, "log_probability", X, min_value=MIN_VALUE)


def test_predict(model, X):
	y_hat = model.predict(X)
	assert_array_almost_equal(y_hat, [0, 1, 1, 1, 0, 1, 0, 1, 0, 1, 0], 4)


def test_predict_raises(model, X):
	_test_raises(model, "predict", X, min_value=MIN_VALUE)

	d = [Exponential(), Exponential()]
	model = BayesClassifier(d)
	_test_raises(model, "predict", X, min_value=MIN_VALUE)


def test_predict_proba(model, X):
	y_hat = model.predict_proba(X)
	assert_array_almost_equal(y_hat,
		[[5.2653e-01, 4.7347e-01],
         [2.6385e-02, 9.7361e-01],
         [1.1551e-07, 1.0000e+00],
         [6.8830e-09, 1.0000e+00],
         [9.7063e-01, 2.9372e-02],
         [1.2660e-15, 1.0000e+00],
         [9.6468e-01, 3.5317e-02],
         [2.3451e-06, 1.0000e+00],
         [9.5759e-01, 4.2413e-02],
         [6.5741e-05, 9.9993e-01],
         [9.9737e-01, 2.6323e-03]], 4)

	model2 = BayesClassifier(model.distributions)
	y_hat2 = model2.predict_proba(X)
	assert_array_almost_equal(y_hat2,
		[[3.2277e-01, 6.7723e-01],
         [1.1481e-02, 9.8852e-01],
         [4.9503e-08, 1.0000e+00],
         [2.9498e-09, 1.0000e+00],
         [9.3405e-01, 6.5951e-02],
         [5.4255e-16, 1.0000e+00],
         [9.2130e-01, 7.8700e-02],
         [1.0050e-06, 1.0000e+00],
         [9.0633e-01, 9.3666e-02],
         [2.8176e-05, 9.9997e-01],
         [9.9388e-01, 6.1207e-03]], 4)	


def test_predict_proba_raises(model, X):
	_test_raises(model, "predict_proba", X, min_value=MIN_VALUE)

	d = [Exponential(), Exponential()]
	model = BayesClassifier(d)
	_test_raises(model, "predict_proba", X, min_value=MIN_VALUE)


def test_predict_log_proba(model, X):
	y_hat = model.predict_log_proba(X)
	assert_array_almost_equal(y_hat,
		[[-6.4145e-01, -7.4766e-01],
         [-3.6350e+00, -2.6740e-02],
         [-1.5974e+01,  0.0000e+00],
         [-1.8794e+01,  0.0000e+00],
         [-2.9812e-02, -3.5277e+00],
         [-3.4303e+01,  0.0000e+00],
         [-3.5955e-02, -3.3434e+00],
         [-1.2963e+01, -2.3842e-06],
         [-4.3338e-02, -3.1603e+00],
         [-9.6298e+00, -6.5804e-05],
         [-2.6357e-03, -5.9399e+00]], 3)

	model2 = BayesClassifier(model.distributions)
	y_hat2 = model2.predict_log_proba(X)
	assert_array_almost_equal(y_hat2,
		[[-1.1308e+00, -3.8974e-01],
         [-4.4671e+00, -1.1548e-02],
         [-1.6821e+01,  0.0000e+00],
         [-1.9642e+01,  0.0000e+00],
         [-6.8226e-02, -2.7188e+00],
         [-3.5150e+01,  0.0000e+00],
         [-8.1969e-02, -2.5421e+00],
         [-1.3810e+01, -9.5367e-07],
         [-9.8348e-02, -2.3680e+00],
         [-1.0477e+01, -2.8133e-05],
         [-6.1395e-03, -5.0961e+00]], 3)	


def test_predict_log_proba_raises(model, X):
	_test_raises(model, "predict_log_proba", X, min_value=MIN_VALUE)

	d = [Exponential(), Exponential()]
	model = BayesClassifier(d)
	_test_raises(model, "predict_log_proba", X, min_value=MIN_VALUE)

###


def test_partial_summarize(model, X, y):
	model.summarize(X[:4], y[:4])
	assert_array_almost_equal(model._w_sum, [2., 2.])

	model.summarize(X[4:], y[4:])
	assert_array_almost_equal(model._w_sum, [6., 5.])


	model = BayesClassifier([Exponential(), Exponential()])
	model.summarize(X[:4], y[:4])
	assert_array_almost_equal(model._w_sum, [2., 2.])

	model.summarize(X[4:], y[4:])
	assert_array_almost_equal(model._w_sum, [6., 5.])


def test_full_summarize(model, X, y):
	model.summarize(X, y)
	assert_array_almost_equal(model._w_sum, [6., 5.])


	model = BayesClassifier([Exponential(), Exponential()])
	model.summarize(X, y)
	assert_array_almost_equal(model._w_sum, [6., 5.])


def test_summarize_weighted(model, X, y, w):
	model.summarize(X, y, sample_weight=w)
	assert_array_almost_equal(model._w_sum, [7., 8.])

	model = BayesClassifier([Exponential(), Exponential()])
	model.summarize(X, y, sample_weight=w)
	assert_array_almost_equal(model._w_sum, [7., 8.])


def test_summarize_weighted_flat(model, X, y, w):
	w = numpy.array(w)[:,0] 

	model.summarize(X, y, sample_weight=w)
	assert_array_almost_equal(model._w_sum, [7., 8.])

	model = BayesClassifier([Exponential(), Exponential()])
	model.summarize(X, y, sample_weight=w)
	assert_array_almost_equal(model._w_sum, [7., 8.])


def test_summarize_weighted_2d(model, X, y):
	model.summarize(X, y, sample_weight=X)
	assert_array_almost_equal(model._w_sum, [7.666667, 5.333333])

	model = BayesClassifier([Exponential(), Exponential()])
	model.summarize(X, y, sample_weight=X)
	assert_array_almost_equal(model._w_sum, [7.666667, 5.333333])


def test_summarize_raises(model, X, y, w):
	assert_raises(ValueError, model.summarize, [X], y)
	assert_raises(ValueError, model.summarize, X[0], y)
	assert_raises((ValueError, TypeError), model.summarize, X[0][0], y)
	assert_raises(ValueError, model.summarize, [x[:-1] for x in X], y)
	assert_raises(ValueError, model.summarize, 
		[[-0.1 for i in range(3)] for x in X], y)

	assert_raises(ValueError, model.summarize, [X], y, w)
	assert_raises(ValueError, model.summarize, X, [y], w)
	assert_raises(ValueError, model.summarize, X, y, [w])
	assert_raises(ValueError, model.summarize, [X], y, [w])
	assert_raises(ValueError, model.summarize, X[:len(X)-1], y, w)
	assert_raises(ValueError, model.summarize, X, y[:len(y)-1], w)
	assert_raises(ValueError, model.summarize, X, y, w[:len(w)-1])


def test_from_summaries(model, X, y):
	model.summarize(X, y)
	model.from_summaries()

	assert_array_almost_equal(model._w_sum, [0., 0.])
	assert_array_almost_equal(model.priors, [6./11, 5./11])
	assert_array_almost_equal(model._log_priors, numpy.log([6./11, 5./11]))

	X_ = numpy.array(X)[numpy.array(y) == 0]
	d = Exponential().fit(X_)

	assert_array_almost_equal(d.scales, model.distributions[0].scales)

	model = BayesClassifier([Exponential(), Exponential()])
	model.summarize(X, y)
	model.from_summaries()

	assert_array_almost_equal(model._w_sum, [0., 0.])
	assert_array_almost_equal(model.priors, [6./11, 5./11])
	assert_array_almost_equal(model._log_priors, numpy.log([6./11, 5./11]))

	assert_array_almost_equal(d.scales, model.distributions[0].scales)


def test_from_summaries_weighted(model, X, y, w):
	model.summarize(X, y, sample_weight=w)
	model.from_summaries()

	assert_array_almost_equal(model._w_sum, [0., 0.])
	assert_array_almost_equal(model.priors, [0.466667, 0.533333])
	assert_array_almost_equal(model._log_priors, 
		numpy.log([0.466667, 0.533333]))

	idxs = numpy.array(y) == 0
	X_ = numpy.array(X)[idxs]
	w_ = numpy.array(w)[idxs]
	d = Exponential().fit(X_, sample_weight=w_)

	assert_array_almost_equal(d.scales, model.distributions[0].scales)

	model = BayesClassifier([Exponential(), Exponential()])
	model.summarize(X, y, sample_weight=w)
	model.from_summaries()

	assert_array_almost_equal(model._w_sum, [0., 0.])
	assert_array_almost_equal(model.priors, [0.466667, 0.533333])
	assert_array_almost_equal(model._log_priors, 
		numpy.log([0.466667, 0.533333]))

	assert_array_almost_equal(d.scales, model.distributions[0].scales)


def test_from_summaries_null(model):
	model.from_summaries()

	assert model.distributions[0].scales[0] != 2.1 
	assert model.distributions[1].scales[0] != 1.5

	assert_array_almost_equal(model._w_sum, [0., 0.])


def test_from_summaries_inertia(X, y):
	d = [Exponential([2.1, 0.3, 0.1]), Exponential([1.5, 3.1, 2.2])]
	model = BayesClassifier(d, priors=[0.2, 0.8], inertia=0.3)
	model.summarize(X, y)
	model.from_summaries()

	assert_array_almost_equal(model._w_sum, [0., 0.])
	assert_array_almost_equal(model.priors, [0.441818, 0.558182])
	assert_array_almost_equal(model._log_priors, 
		numpy.log([0.441818, 0.558182]))


	d = [Exponential([2.1, 0.3, 0.1]), Exponential([1.5, 3.1, 2.2])]
	model = BayesClassifier(d, priors=[0.2, 0.8], inertia=1.0)
	model.summarize(X, y)
	model.from_summaries()

	assert_array_almost_equal(model._w_sum, [0., 0.])
	assert_array_almost_equal(model.priors, [0.2, 0.8])
	assert_array_almost_equal(model._log_priors, numpy.log([0.2, 0.8]))


	s1, s2 = [2.1, 0.3, 0.1], [1.5, 3.1, 2.2] 
	d = [Exponential(s1, inertia=1.0), Exponential(s2, inertia=1.0)]
	model = BayesClassifier(d, priors=[0.2, 0.8])
	model.summarize(X, y)
	model.from_summaries()

	assert_array_almost_equal(model.distributions[0].scales, s1)
	assert_array_almost_equal(model.distributions[1].scales, s2)

	assert_array_almost_equal(model._w_sum, [0., 0.])
	assert_array_almost_equal(model.priors, [0.545455, 0.454545])
	assert_array_almost_equal(model._log_priors, 
		numpy.log([0.545455, 0.454545]))


	d = [Exponential(s1, inertia=1.0), Exponential(s2)]
	model = BayesClassifier(d, priors=[0.2, 0.8], inertia=1.0)
	model.summarize(X, y)
	model.from_summaries()

	assert_array_almost_equal(model.distributions[0].scales, s1)
	assert_array_almost_equal(model.distributions[1].scales, [1.2, 1.4, 0.6])

	assert_array_almost_equal(model._w_sum, [0., 0.])
	assert_array_almost_equal(model.priors, [0.2, 0.8])
	assert_array_almost_equal(model._log_priors, numpy.log([0.2, 0.8]))


def test_from_summaries_weighted_inertia(X, y, w):
	d = [Exponential([2.1, 0.3, 0.1]), Exponential([1.5, 3.1, 2.2])]
	model = BayesClassifier(d, priors=[0.2, 0.8], inertia=0.3)
	model.summarize(X, y, sample_weight=w)
	model.from_summaries()

	assert_array_almost_equal(model._w_sum, [0., 0.])
	assert_array_almost_equal(model.priors, [0.386667, 0.613333])
	assert_array_almost_equal(model._log_priors, 
		numpy.log([0.386667, 0.613333]))


	d = [Exponential([2.1, 0.3, 0.1]), Exponential([1.5, 3.1, 2.2])]
	model = BayesClassifier(d, priors=[0.2, 0.8], inertia=1.0)
	model.summarize(X, y, sample_weight=w)
	model.from_summaries()

	assert_array_almost_equal(model._w_sum, [0., 0.])
	assert_array_almost_equal(model.priors, [0.2, 0.8])
	assert_array_almost_equal(model._log_priors, numpy.log([0.2, 0.8]))


def test_from_summaries_frozen(model, X, y):
	d = [Exponential([2.1, 0.3, 0.1]), Exponential([1.5, 3.1, 2.2])]
	model = BayesClassifier(d, priors=[0.2, 0.8], frozen=True)
	model.summarize(X, y)
	model.from_summaries()

	assert_array_almost_equal(model._w_sum, [0., 0.])
	assert_array_almost_equal(model.priors, [0.2, 0.8])
	assert_array_almost_equal(model._log_priors, numpy.log([0.2, 0.8]))


	s1, s2 = [2.1, 0.3, 0.1], [1.5, 3.1, 2.2] 
	d = [Exponential(s1, frozen=True), Exponential(s2, frozen=True)]
	model = BayesClassifier(d, priors=[0.2, 0.8])
	model.summarize(X, y)
	model.from_summaries()

	assert_array_almost_equal(model.distributions[0].scales, s1)
	assert_array_almost_equal(model.distributions[1].scales, s2)

	assert_array_almost_equal(model._w_sum, [0., 0.])
	assert_array_almost_equal(model.priors, [0.545455, 0.454545])
	assert_array_almost_equal(model._log_priors, 
		numpy.log([0.545455, 0.454545]))


	d = [Exponential(s1, frozen=True), Exponential(s2)]
	model = BayesClassifier(d, priors=[0.2, 0.8], frozen=True)
	model.summarize(X, y)
	model.from_summaries()

	assert_array_almost_equal(model.distributions[0].scales, s1)
	assert_array_almost_equal(model.distributions[1].scales, [1.2, 1.4, 0.6])

	assert_array_almost_equal(model._w_sum, [0., 0.])
	assert_array_almost_equal(model.priors, [0.2, 0.8])
	assert_array_almost_equal(model._log_priors, numpy.log([0.2, 0.8]))


def test_fit(model, X, y):
	model.fit(X, y)

	assert_array_almost_equal(model._w_sum, [0., 0.])
	assert_array_almost_equal(model.priors, [6./11, 5./11])
	assert_array_almost_equal(model._log_priors, numpy.log([6./11, 5./11]))

	X_ = numpy.array(X)[numpy.array(y) == 0]
	d = Exponential().fit(X_)

	assert_array_almost_equal(d.scales, model.distributions[0].scales)

	model = BayesClassifier([Exponential(), Exponential()])
	model.fit(X, y)

	assert_array_almost_equal(model._w_sum, [0., 0.])
	assert_array_almost_equal(model.priors, [6./11, 5./11])
	assert_array_almost_equal(model._log_priors, numpy.log([6./11, 5./11]))

	assert_array_almost_equal(d.scales, model.distributions[0].scales)


def test_fit_weighted(model, X, y, w):
	model.fit(X, y, sample_weight=w)

	assert_array_almost_equal(model._w_sum, [0., 0.])
	assert_array_almost_equal(model.priors, [0.466667, 0.533333])
	assert_array_almost_equal(model._log_priors, 
		numpy.log([0.466667, 0.533333]))

	idxs = numpy.array(y) == 0
	X_ = numpy.array(X)[idxs]
	w_ = numpy.array(w)[idxs]
	d = Exponential().fit(X_, sample_weight=w_)

	assert_array_almost_equal(d.scales, model.distributions[0].scales)

	model = BayesClassifier([Exponential(), Exponential()])
	model.fit(X, y, sample_weight=w)

	assert_array_almost_equal(model._w_sum, [0., 0.])
	assert_array_almost_equal(model.priors, [0.466667, 0.533333])
	assert_array_almost_equal(model._log_priors, 
		numpy.log([0.466667, 0.533333]))

	assert_array_almost_equal(d.scales, model.distributions[0].scales)


def test_fit_chain(X, y):
	model = BayesClassifier([Exponential(), Exponential()]).fit(X, y)

	assert_array_almost_equal(model._w_sum, [0., 0.])
	assert_array_almost_equal(model.priors, [6./11, 5./11])
	assert_array_almost_equal(model._log_priors, numpy.log([6./11, 5./11]))


def test_fit_raises(model, X, w, y):
	assert_raises(ValueError, model.fit, [X], y)
	assert_raises(ValueError, model.fit, X[0], y)
	assert_raises((ValueError, TypeError), model.fit, X[0][0], y)
	assert_raises(ValueError, model.fit, [x[:-1] for x in X], y)
	assert_raises(ValueError, model.fit, 
		[[-0.1 for i in range(3)] for x in X], y)

	assert_raises(ValueError, model.fit, [X], y, w)
	assert_raises(ValueError, model.fit, X, [y], w)
	assert_raises(ValueError, model.fit, X, y, [w])
	assert_raises(ValueError, model.fit, [X], y, [w])
	assert_raises(ValueError, model.fit, X[:len(X)-1], y, w)
	assert_raises(ValueError, model.fit, X, y[:len(y)-1], w)
	assert_raises(ValueError, model.fit, X, y, w[:len(w)-1])


def test_serialization(X, model):
	torch.save(model, ".pytest.torch")
	model2 = torch.load(".pytest.torch")
	os.system("rm .pytest.torch")

	assert_array_almost_equal(model2.priors, model.priors)
	assert_array_almost_equal(model2._log_priors, model._log_priors)

	assert_array_almost_equal(model2.predict_proba(X), model.predict_proba(X))

	m1d1, m1d2 = model.distributions
	m2d1, m2d2 = model2.distributions

	assert m1d1 is not m2d1
	assert m1d2 is not m2d2

	assert_array_almost_equal(m1d1.scales, m2d1.scales)
	assert_array_almost_equal(m1d2.scales, m2d2.scales)


def test_masked_probability(model, X, X_masked):
	X = torch.tensor(numpy.array(X))
	y = [1.668138e-02, 1.911842e-02, 4.393471e-03, 1.633741e-03,
           9.786682e-02, 1.229918e-04, 1.585297e-01, 6.066021e-03,
           2.571130e-01, 9.765117e-03, 1.114044e+01]

	mask = torch.ones_like(X).type(torch.bool)
	X_ = torch.masked.MaskedTensor(X, mask=mask)

	assert_array_almost_equal(y, model.probability(X_), 5)

	y =  [5.277007e-02, 1.175627e+00, 1.000000e+00, 1.633741e-03,
           1.533307e-01, 1.229918e-04, 1.000000e+00, 1.880462e-02,
           2.571130e-01, 9.765117e-03, 3.424242e+00]

	assert_array_almost_equal(y, model.probability(X_masked), 5)


def test_masked_log_probability(model, X, X_masked):
	X = torch.tensor(numpy.array(X))
	y = [-4.09346, -3.9571 , -5.42764, -6.41688, -2.32415, -9.00339,
           -1.84181, -5.10505, -1.35824, -4.62894,  2.41058]

	mask = torch.ones_like(X).type(torch.bool)
	X_ = torch.masked.MaskedTensor(X, mask=mask)

	assert_array_almost_equal(y, model.log_probability(X_), 5)

	y = [-2.94181,  0.1618 ,  0.     , -6.41688, -1.87516, -9.00339,
            0.     , -3.97365, -1.35824, -4.62894,  1.23088]

	assert_array_almost_equal(y, model.log_probability(X_masked), 5)


def test_masked_emission_matrix(model, X, X_masked):
	X = torch.tensor(numpy.array(X))
	mask = torch.ones_like(X).type(torch.bool)
	X_ = torch.masked.MaskedTensor(X, mask=mask)

	e = model._emission_matrix(X_)

	assert_array_almost_equal(e, 
		[[ -4.7349,  -4.8411],
         [ -7.5921,  -3.9838],
         [-21.4016,  -5.4276],
         [-25.2111,  -6.4169],
         [ -2.3540,  -5.8519],
         [-43.3063,  -9.0034],
         [ -1.8778,  -5.1852],
         [-18.0682,  -5.1051],
         [ -1.4016,  -4.5185],
         [-14.2587,  -4.6290],
         [  2.4079,  -3.5293]], 4)

	d = [Exponential([2.1, 0.3, 0.1]), Exponential([1.5, 3.1, 2.2])]
	model = BayesClassifier(d)
	e = model._emission_matrix(X_masked)

	assert_array_almost_equal(e, 
		[[ -3.8533,  -3.2582],
         [ -0.2311,  -2.2300],
         [ -0.6931,  -0.6931],
         [-25.5476,  -5.9061],
         [ -2.8225,  -2.1471],
         [-43.6428,  -8.4926],
         [ -0.6931,  -0.6931],
         [-19.6087,  -3.4628],
         [ -1.7381,  -4.0077],
         [-14.5952,  -4.1182],
         [  0.8675,  -1.8871]], 4)


def test_masked_summarize(model, X, X_masked, w, y):
	X = torch.tensor(numpy.array(X))
	mask = torch.ones_like(X).type(torch.bool)
	X_ = torch.masked.MaskedTensor(X, mask=mask)

	d = [Exponential([2.1, 0.3, 0.1]), Exponential([1.5, 3.1, 2.2])]
	model = BayesClassifier(d, priors=[0.7, 0.3])
	model.summarize(X_, y, sample_weight=w)
	assert_array_almost_equal(model._w_sum, [7.0, 8.0])

	d = [Exponential([2.1, 0.3, 0.1]), Exponential([1.5, 3.1, 2.2])]
	model = BayesClassifier(d, priors=[0.7, 0.3])
	model.summarize(X_masked, y, sample_weight=w)
	assert_array_almost_equal(model._w_sum, [5.0, 8.0])


def test_masked_from_summaries(model, X, X_masked, y):
	X = torch.tensor(numpy.array(X))
	mask = torch.ones_like(X).type(torch.bool)
	X_ = torch.masked.MaskedTensor(X, mask=mask)

	model.summarize(X_, y)
	model.from_summaries()

	assert_array_almost_equal(model._w_sum, [0., 0.])
	assert_array_almost_equal(model.priors, [0.545455, 0.454545])
	assert_array_almost_equal(model._log_priors, 
		numpy.log([0.545455, 0.454545]))

	d = [Exponential([2.1, 0.3, 0.1]), Exponential([1.5, 3.1, 2.2])]
	model = BayesClassifier(d, priors=[0.7, 0.3])
	model.summarize(X_masked, y, sample_weight=w)
	model.from_summaries()

	assert_array_almost_equal(model._w_sum, [0., 0.])
	assert_array_almost_equal(model.priors, [0.444444, 0.555556])
	assert_array_almost_equal(model._log_priors, 
		numpy.log([0.444444, 0.555556]))


def test_masked_fit(X, X_masked, y):
	X = torch.tensor(numpy.array(X))
	mask = torch.ones_like(X).type(torch.bool)
	X_ = torch.masked.MaskedTensor(X, mask=mask)

	d = [Exponential([2.1, 0.3, 1.1]), Exponential([1.5, 3.1, 2.2])]
	model = BayesClassifier(d)
	model.fit(X_, y)

	assert_array_almost_equal(model._w_sum, [0., 0.])
	assert_array_almost_equal(model.priors, [0.545455, 0.454545])
	assert_array_almost_equal(model._log_priors, 
		numpy.log([0.545455, 0.454545]))

	d = [Exponential([2.1, 0.3, 0.1]), Exponential([1.5, 3.1, 2.2])]
	model = BayesClassifier(d)
	model.fit(X_masked, y)

	assert_array_almost_equal(model._w_sum, [0., 0.])
	assert_array_almost_equal(model.priors, [0.444444, 0.555556])
	assert_array_almost_equal(model._log_priors, 
		numpy.log([0.444444, 0.555556]))


# conditional_categorical.py
# Contact: Jacob Schreiber <jmschreiber91@gmail.com>

import numpy
import torch
import itertools

from .._utils import _cast_as_tensor
from .._utils import _cast_as_parameter
from .._utils import _update_parameter
from .._utils import _check_parameter
from .._utils import _reshape_weights

from .._utils import BufferList

from ._distribution import ConditionalDistribution
from .categorical import Categorical


class ConditionalCategorical(ConditionalDistribution):
	"""A conditional categorical distribution.

	This is a categorical distribution that is conditioned on previous
	emissions, meaning that the probability of each character depends on the
	observed character earlier in the sequence. Each feature is conditioned
	independently of the others like a `Categorical` distribution. 

	This conditioning makes the shape of the distribution a bit more
	complicated than the `JointCategorical` distribution. Specifically, a 
	`JointCategorical` distribution is multivariate by definition but a
	`ConditionalCategorical` does not have to be. Although both may appear 
	similar in that they both take in a vector of characters and return 
	probabilities, the vector fed into the JointCategorical are all observed 
	together without some notion of time, whereas the ConditionalCategorical 
	explicitly requires a notion of timing, where the probability of later 
	characters depend on the composition of characters seen before.


	Parameters
	----------
	probs: list of numpy.ndarray, torch.tensor or None, shape=(k, k), optional
		A list of conditional probabilities with one tensor for each feature
		in the data being modeled. Each tensor should have `k+1` dimensions 
		where `k` is the number of timesteps to condition on. Each dimension
		should span the number of keys in that dimension. For example, if
		specifying a univariate conditional categorical distribution where
		k=2, a valid tensor shape would be [(2, 3, 4)]. Default is None.

	n_categories: list, numpy.ndarray, torch.tensor or None, optional
		The number of categories for each feature in the data. Only needs to
		be provided when the parameters will be learned directly from data and
		you want to make sure that right number of keys are included in each
		dimension. Unlike the `Categorical` distribution, this needs to be
		a list of shapes with one shape for each feature and the shape matches
		that specified in `probs`. Default is None.

	pseudocount: float, optional
		A value to add to the observed counts of each feature when training.
		Setting this to a positive value ensures that no probabilities are
		truly zero. Default is 0.

	inertia: float, (0, 1), optional
		Indicates the proportion of the update to apply to the parameters
		during training. When the inertia is 0.0, the update is applied in
		its entirety and the previous parameters are ignored. When the
		inertia is 1.0, the update is entirely ignored and the previous
		parameters are kept, equivalently to if the parameters were frozen.

	frozen: bool, optional
		Whether all the parameters associated with this distribution are frozen.
		If you want to freeze individual pameters, or individual values in those
		parameters, you must modify the `frozen` attribute of the tensor or
		parameter directly. Default is False.

	check_data: bool, optional
		Whether to check properties of the data and potentially recast it to
		torch.tensors. This does not prevent checking of parameters but can
		slightly speed up computation when you know that your inputs are valid.
		Setting this to False is also necessary for compiling.
	"""

	def __init__(self, probs=None, n_categories=None, pseudocount=0, 
		inertia=0.0, frozen=False, check_data=True):
		super().__init__(inertia=inertia, frozen=frozen, check_data=check_data)
		self.name = "ConditionalCategorical"

		if probs is not None:
			self.n_categories = []
			self.probs = torch.nn.ParameterList([])
			
			for prob in probs:
				prob = _check_parameter(_cast_as_parameter(prob), "probs",
					min_value=0, max_value=1)
				
				self.probs.append(prob)
				self.n_categories.append(tuple(prob.shape))

		else:
			self.probs = None
			self.n_categories = n_categories
		
		self.pseudocount = _check_parameter(pseudocount, "pseudocount")

		self._initialized = probs is not None
		self.d = len(self.probs) if self._initialized else None
		self.n_parents = len(self.probs[0].shape) if self._initialized else None
		self._reset_cache()

	def _initialize(self, d, n_categories):
		"""Initialize the probability distribution.

		This method is meant to only be called internally. It initializes the
		parameters of the distribution and stores its dimensionality. For more
		complex methods, this function will do more.


		Parameters
		----------
		d: int
			The dimensionality the distribution is being initialized to.

		n_categories: list of tuples
			The shape of each conditional distribution, one per feature.
		"""

		self.n_categories = []
		for n_cat in n_categories:
			if isinstance(n_cat, (list, tuple)):
				self.n_categories.append(tuple(n_cat))
			elif isinstance(n_cat, (numpy.ndarray, torch.Tensor)):
				self.n_categories.append(tuple(n_cat.tolist()))

		self.n_parents = len(self.n_categories[0])
		self.probs = torch.nn.ParameterList([_cast_as_parameter(torch.zeros(
			*cats, dtype=self.dtype, device=self.device, requires_grad=False)) 
				for cats in self.n_categories])

		self._initialized = True
		super()._initialize(d)

	def _reset_cache(self):
		"""Reset the internally stored statistics.

		This method is meant to only be called internally. It resets the
		stored statistics used to update the model parameters as well as
		recalculates the cached values meant to speed up log probability
		calculations.
		"""

		if self._initialized == False:
			return

		_w_sum = []
		_xw_sum = []

		for n_categories in self.n_categories:
			_w_sum.append(torch.zeros(*n_categories[:-1], 
				dtype=self.probs[0].dtype, device=self.device))
			_xw_sum.append(torch.zeros(*n_categories, 
				dtype=self.probs[0].dtype, device=self.device))

		self._w_sum = BufferList(_w_sum)
		self._xw_sum = BufferList(_xw_sum)

		self._log_probs = BufferList([torch.log(prob) for prob in self.probs])

	def sample(self, n, X):
		"""Sample from the probability distribution.

		This method will return `n` samples generated from the underlying
		probability distribution. For a mixture model, this involves first
		sampling the component using the prior probabilities, and then sampling
		from the chosen distribution.


		Parameters
		----------
		n: int
			The number of samples to generate.
		
		X: list, numpy.ndarray, torch.tensor, shape=(n, d, *self.probs.shape-1) 
			The values to be conditioned on when generating the samples.

		Returns
		-------
		X: torch.tensor, shape=(n, self.d)
			Randomly generated samples.
		"""

		X = _check_parameter(_cast_as_tensor(X), "X", ndim=3, 
			shape=(-1, self.n_parents-1, self.d))

		y = []
		for i in range(n):
			y.append([])

			for j in range(self.d):
				idx = tuple(X[i, :, j])
				if len(idx) == 1:
					idx = idx[0].item()
				
				probs = self.probs[j][idx]

				y_ = torch.multinomial(probs, 1).item()
				y[-1].append(y_)

		return torch.tensor(y)

	def log_probability(self, X):
		X = _check_parameter(_cast_as_tensor(X), "X", ndim=3, 
			shape=(-1, self.n_parents, self.d), check_parameter=self.check_data)

		logps = torch.zeros(len(X), dtype=self.probs[0].dtype, device=X.device, 
			requires_grad=False)

		for i in range(len(X)):
			for j in range(self.d):
				logps[i] += self._log_probs[j][tuple(X[i, :, j])]

		return logps

	def summarize(self, X, sample_weight=None):
		if self.frozen == True:
			return

		X = _check_parameter(_cast_as_tensor(X), "X", ndim=3, 
			dtypes=(torch.int32, torch.int64), check_parameter=self.check_data)

		if not self._initialized:
			self._initialize(len(X[0][0]), torch.max(X, dim=0)[0].T+1)

		X = _check_parameter(X, "X", shape=(-1, self.n_parents, self.d),
			check_parameter=self.check_data)
		sample_weight = _check_parameter(_cast_as_tensor(sample_weight, 
			dtype=torch.float32), "sample_weight", min_value=0, ndim=(1, 2))

		if sample_weight is None:
			sample_weight = torch.ones(X[:, 0].shape[0], X[:, 0].shape[-1], 
				dtype=self.probs[0].dtype)
		elif len(sample_weight.shape) == 1: 
			sample_weight = sample_weight.reshape(-1, 1).expand(-1, X.shape[2])
		elif sample_weight.shape[1] == 1 and self.d > 1:
			sample_weight = sample_weight.expand(-1, X.shape[2])

		_check_parameter(sample_weight, "sample_weight", 
			min_value=0, ndim=2, shape=(X.shape[0], X.shape[2]))

		for j in range(self.d):
			strides = torch.tensor(self._xw_sum[j].stride(), device=X.device)
			X_ = torch.sum(X[:, :, j] * strides, dim=-1)

			self._xw_sum[j].view(-1).scatter_add_(0, X_, sample_weight[:,j])
			self._w_sum[j][:] = self._xw_sum[j].sum(dim=-1)

	def from_summaries(self):
		if self.frozen == True:
			return

		for i in range(self.d):
			probs = self._xw_sum[i] / self._w_sum[i].unsqueeze(-1)
			probs = torch.nan_to_num(probs, 1. / probs.shape[-1])

			_update_parameter(self.probs[i], probs, self.inertia)

		self._reset_cache()




# zero_inflated.py
# Contact: Jacob Schreiber <jmschreiber91@gmail.com>

import time
import torch

from .._utils import _cast_as_tensor
from .._utils import _cast_as_parameter
from .._utils import _update_parameter
from .._utils import _check_parameter
from .._utils import _reshape_weights

from ._distribution import Distribution


class ZeroInflated(Distribution):
	"""A wrapper for a zero-inflated distribution.

	Some discrete distributions, e.g. Poisson or negative binomial, are used
	to model data that has many more zeroes in it than one would expect from
	the true signal itself. Potentially, this is because data collection devices
	fail or other gaps exist in the data. A zero-inflated distribution is
	essentially a mixture of these zero values and the real underlying
	distribution.

	Accordingly, this class serves as a wrapper that can be dropped in for
	other probability distributions and makes them "zero-inflated". It is
	similar to a mixture model between the distribution passed in and a dirac
	delta distribution, except that the mixture happens independently for each
	distribution as well as for each example.


	Parameters
	----------
	distribution: pomegranate.distributions.Distribution
		A pomegranate distribution object. It should probably be a discrete
		distribution, but does not technically have to be.

	priors: tuple, numpy.ndarray, torch.Tensor, or None. shape=(2,), optional
		The prior probabilities over the given distribution and the dirac
		delta component. Default is None.

	max_iter: int, optional
		The number of iterations to do in the EM step of fitting the
		distribution. Default is 10.

	tol: float, optional
		The threshold at which to stop during fitting when the improvement
		goes under. Default is 0.1.

	inertia: float, [0, 1], optional
		Indicates the proportion of the update to apply to the parameters
		during training. When the inertia is 0.0, the update is applied in
		its entirety and the previous parameters are ignored. When the
		inertia is 1.0, the update is entirely ignored and the previous
		parameters are kept, equivalently to if the parameters were frozen.

	frozen: bool, optional
		Whether all the parameters associated with this distribution are frozen.
		If you want to freeze individual pameters, or individual values in those
		parameters, you must modify the `frozen` attribute of the tensor or
		parameter directly. Default is False.

	verbose: bool, optional
		Whether to print the improvement and timings during training.
	"""

	def __init__(self, distribution, priors=None, max_iter=10, 
		tol=0.1, inertia=0.0, frozen=False, check_data=False, verbose=False):
		super().__init__(inertia=inertia, frozen=frozen, check_data=check_data)
		self.name = "ZeroInflated"

		self.distribution = distribution
		self.priors = _check_parameter(_cast_as_parameter(priors), "priors", 
			min_value=0, max_value=1, ndim=1, value_sum=1.0)

		self.verbose = verbose
		self._initialized = distribution._initialized is True
		self.d = distribution.d if self._initialized else None
		
		self.max_iter = max_iter
		self.tol = tol

		if self.priors is None and self.d is not None:
			self.priors = _cast_as_parameter(torch.ones(self.d, 
				device=self.device) / 2)

		self._reset_cache()

	def _initialize(self, X):
		"""Initialize the probability distribution.

		This method is meant to only be called internally. It initializes the
		parameters of the distribution and stores its dimensionality. For more
		complex methods, this function will do more.


		Parameters
		----------
		X: list, numpy.ndarray, torch.Tensor, shape=(1, self.d)
			The data to use to initialize the model.
		"""

		self.distribution._initialize(X.shape[1])
		self.distribution.fit(X)

		self.priors = _cast_as_parameter(torch.ones(X.shape[1], 
			device=self.device) / 2)
		self._initialized = True
		super()._initialize(X.shape[1])

	def _reset_cache(self):
		"""Reset the internally stored statistics.

		This method is meant to only be called internally. It resets the
		stored statistics used to update the model parameters as well as
		recalculates the cached values meant to speed up log probability
		calculations.
		"""

		if self._initialized == False:
			return

		self.register_buffer("_w_sum", torch.zeros(self.d, 2, 
			device=self.device))
		self.register_buffer("_log_priors", torch.log(self.priors))

	def _emission_matrix(self, X):
		"""Return the emission/responsibility matrix.

		This method returns the log probability of each example under each
		distribution contained in the model with the log prior probability
		of each component added.


		Parameters
		----------
		X: list, tuple, numpy.ndarray, torch.Tensor, shape=(-1, self.d)
			A set of examples to evaluate. 

	
		Returns
		-------
		e: torch.Tensor, shape=(-1, self.k)
			A set of log probabilities for each example under each distribution.
		"""

		X = _check_parameter(_cast_as_tensor(X), "X", ndim=2, 
			shape=(-1, self.d))

		e = torch.empty(X.shape[0], self.d, 2, device=self.device)
		e[:, :, 0] = self._log_priors.unsqueeze(0)
		e[:, :, 0] += self.distribution.log_probability(X).unsqueeze(1)
		
		e[:, :, 1] = torch.log(1 - self.priors).unsqueeze(0)
		e[:, :, 1] += torch.where(X == 0, 0, float("-inf"))
		return e

	def fit(self, X, sample_weight=None):
		"""Fit the model to optionally weighted examples.

		This method implements the core of the learning process. For a
		zero-inflated distribution, this involves performing EM until the
		distribution being fit converges.

		This method is largely a wrapper around the `summarize` and
		`from_summaries` methods. It's primary contribution is serving as a
		loop around these functions and to monitor convergence.


		Parameters
		----------
		X: list, tuple, numpy.ndarray, torch.Tensor, shape=(-1, self.d)
			A set of examples to evaluate. 

		sample_weight: list, tuple, numpy.ndarray, torch.Tensor, optional
			A set of weights for the examples. This can be either of shape
			(-1, self.d) or a vector of shape (-1,). Default is ones.


		Returns
		-------
		self
		"""		

		logp = None
		for i in range(self.max_iter):
			start_time = time.time()

			last_logp = logp
			logp = self.summarize(X, sample_weight=sample_weight)

			if i > 0:
				improvement = logp - last_logp
				duration = time.time() - start_time

				if self.verbose:
					print("[{}] Improvement: {}, Time: {:4.4}s".format(i, 
						improvement, duration))

				if improvement < self.tol:
					break

			self.from_summaries()

		self._reset_cache()
		return self

	def summarize(self, X, sample_weight=None):
		"""Extract the sufficient statistics from a batch of data.

		This method calculates the sufficient statistics from optionally
		weighted data and adds them to the stored cache. The examples must be
		given in a 2D format. Sample weights can either be provided as one
		value per example or as a 2D matrix of weights for each feature in
		each example.


		Parameters
		----------
		X: list, tuple, numpy.ndarray, torch.Tensor, shape=(-1, self.d)
			A set of examples to summarize.

		sample_weight: list, tuple, numpy.ndarray, torch.Tensor, optional
			A set of weights for the examples. This can be either of shape
			(-1, self.d) or a vector of shape (-1,). Default is ones.
		"""

		X = _cast_as_tensor(X)
		if not self._initialized:
			self._initialize(X)

		_check_parameter(X, "X", ndim=2, shape=(-1, self.d))
		sample_weight = _reshape_weights(X, _cast_as_tensor(sample_weight, 
			dtype=torch.float32), device=self.device)

		e = self._emission_matrix(X)
		logp = torch.logsumexp(e, dim=2, keepdims=True)
		y = torch.exp(e - logp)

		self.distribution.summarize(X, y[:, :, 0] * sample_weight)

		if not self.frozen:
			self._w_sum += torch.sum(y * sample_weight.unsqueeze(-1), dim=(0, 1)) 

		return torch.sum(logp)

	def from_summaries(self):
		"""Update the model parameters given the extracted statistics.

		This method uses calculated statistics from calls to the `summarize`
		method to update the distribution parameters. Hyperparameters for the
		update are passed in at initialization time.

		Note: Internally, a call to `fit` is just a successive call to the
		`summarize` method followed by the `from_summaries` method.
		"""

		self.distribution.from_summaries()

		if self.frozen == True:
			return

		priors = self._w_sum[:,0] / torch.sum(self._w_sum, dim=-1)

		_update_parameter(self.priors, priors, self.inertia)
		self._reset_cache()



# gmm.py
# Author: Jacob Schreiber <jmschreiber91@gmail.com>

import time
import numpy
import torch

from ._utils import _cast_as_tensor
from ._utils import _cast_as_parameter
from ._utils import _update_parameter
from ._utils import _check_parameter
from ._utils import _reshape_weights

from .distributions._distribution import Distribution

from ._bayes import BayesMixin

from .kmeans import KMeans


class GeneralMixtureModel(BayesMixin, Distribution):
	"""A general mixture model.

	Frequently, data is generated from multiple components. A mixture model
	is a probabilistic model that explicitly models data as having come from
	a set of probability distributions rather than a single one. Usually, the
	abbreviation "GMM" refers to a Gaussian mixture model, but any probability
	distribution or heterogeneous set of distributions can be included in the
	mixture, making it a "general" mixture model.

	However, a mixture model itself has all the same theoretical properties as a
	probability distribution because it is one. Hence, it can be used in any
	situation that a simpler distribution could, such as an emission
	distribution for a HMM or a component of a Bayes classifier.

	Conversely, many models that are usually thought of as composed of
	probability distributions but distinct from them, e.g. hidden Markov models,
	Markov chains, and Bayesian networks, can in theory be passed into this
	object and incorporated into the mixture.

	If the distributions included in the mixture are not initialized, the
	fitting step will first initialize them by running k-means for a small
	number of iterations and fitting the distributions to the clusters that
	are discovered.


	Parameters
	----------
	distributions: tuple or list
		A set of distribution objects. These objects do not need to be
		initialized, i.e., can be "Normal()". 

	priors: tuple, numpy.ndarray, torch.Tensor, or None. shape=(k,), optional
		The prior probabilities over the given distributions. Default is None.

	max_iter: int, optional
		The number of iterations to do in the EM step of fitting the
		distribution. Default is 10.

	tol: float, optional
		The threshold at which to stop during fitting when the improvement
		goes under. Default is 0.1.

	inertia: float, [0, 1], optional
		Indicates the proportion of the update to apply to the parameters
		during training. When the inertia is 0.0, the update is applied in
		its entirety and the previous parameters are ignored. When the
		inertia is 1.0, the update is entirely ignored and the previous
		parameters are kept, equivalently to if the parameters were frozen.

	frozen: bool, optional
		Whether all the parameters associated with this distribution are frozen.
		If you want to freeze individual pameters, or individual values in those
		parameters, you must modify the `frozen` attribute of the tensor or
		parameter directly. Default is False.

	check_data: bool, optional
		Whether to check properties of the data and potentially recast it to
		torch.tensors. This does not prevent checking of parameters but can
		slightly speed up computation when you know that your inputs are valid.
		Setting this to False is also necessary for compiling. Default is True.

	verbose: bool, optional
		Whether to print the improvement and timings during training.
	"""

	def __init__(self, distributions, priors=None, init='random', max_iter=1000, 
		tol=0.1, inertia=0.0, frozen=False, random_state=None, check_data=True, 
		verbose=False):
		super().__init__(inertia=inertia, frozen=frozen, check_data=check_data)
		self.name = "GeneralMixtureModel"

		_check_parameter(distributions, "distributions", dtypes=(list, tuple, 
			numpy.array, torch.nn.ModuleList))
		self.distributions = torch.nn.ModuleList(distributions)

		self.priors = _check_parameter(_cast_as_parameter(priors), "priors", 
			min_value=0, max_value=1, ndim=1, value_sum=1.0, 
			shape=(len(distributions),))

		self.verbose = verbose

		self.k = len(distributions)

		if all(d._initialized for d in distributions):
			self._initialized = True
			self.d = distributions[0].d
			if self.priors is None:
				self.priors = _cast_as_parameter(torch.ones(self.k) / self.k)

		else:
			self._initialized = False
			self.d = None
		
		self.init = init
		self.max_iter = max_iter
		self.tol = tol
		self.random_state = random_state
		self._reset_cache()

	def _initialize(self, X, sample_weight=None):
		"""Initialize the probability distribution.

		This method is meant to only be called internally. It initializes the
		parameters of the distribution and stores its dimensionality. For more
		complex methods, this function will do more.


		Parameters
		----------
		X: list, numpy.ndarray, torch.Tensor, shape=(-1, self.d)
			The data to use to initialize the model.

		sample_weight: list, tuple, numpy.ndarray, torch.Tensor, optional
			A set of weights for the examples. This can be either of shape
			(-1, self.d) or a vector of shape (-1,). Default is ones.
		"""

		X = _check_parameter(_cast_as_tensor(X), "X", ndim=2)

		if sample_weight is None:
			sample_weight = torch.ones(1, dtype=self.dtype, 
				device=self.device).expand(X.shape[0], 1)
		else:
			sample_weight = _check_parameter(_cast_as_tensor(sample_weight), 
				"sample_weight", min_value=0., check_parameter=self.check_data)

		model = KMeans(self.k, init=self.init, max_iter=3, 
			random_state=self.random_state)

		if self.device != model.device:
			model.to(self.device)

		y_hat = model.fit_predict(X, sample_weight=sample_weight)

		self.priors = _cast_as_parameter(torch.empty(self.k, 
			dtype=self.dtype, device=self.device))

		sample_weight_sum = sample_weight.sum()
		for i in range(self.k):
			idx = y_hat == i

			sample_weight_idx = sample_weight[idx]
			self.distributions[i].fit(X[idx], sample_weight=sample_weight_idx)
			self.priors[i] = sample_weight_idx.sum() / sample_weight_sum

		self._initialized = True
		self._reset_cache()
		super()._initialize(X.shape[1])

	def sample(self, n):
		"""Sample from the probability distribution.

		This method will return `n` samples generated from the underlying
		probability distribution. For a mixture model, this involves first
		sampling the component using the prior probabilities, and then sampling
		from the chosen distribution.


		Parameters
		----------
		n: int
			The number of samples to generate.
		

		Returns
		-------
		X: torch.tensor, shape=(n, self.d)
			Randomly generated samples.
		"""

		X = []
		for distribution in self.distributions:
			X_ = distribution.sample(n)
			X.append(X_)

		X = torch.stack(X)
		idxs = torch.multinomial(self.priors, num_samples=n, replacement=True)
		return X[idxs, torch.arange(n)]

	def fit(self, X, sample_weight=None, priors=None):
		"""Fit the model to optionally weighted examples.

		This method implements the core of the learning process. For a
		mixture model, this involves performing EM until the distributions that
		are being fit converge according to the threshold set by `tol`, or
		until the maximum number of iterations has been hit.

		This method is largely a wrapper around the `summarize` and
		`from_summaries` methods. It's primary contribution is serving as a
		loop around these functions and to monitor convergence.


		Parameters
		----------
		X: list, tuple, numpy.ndarray, torch.Tensor, shape=(-1, self.d)
			A set of examples to evaluate. 

		sample_weight: list, tuple, numpy.ndarray, torch.Tensor optional
			A set of weights for the examples. This can be either of shape
			(-1, self.d) or a vector of shape (-1,). Default is ones.

		priors: list, numpy.ndarray, torch.Tensor, shape=(-1, self.k)
			Prior probabilities of assigning each symbol to each node. If not
			provided, do not include in the calculations (conceptually
			equivalent to a uniform probability, but without scaling the
			probabilities). This can be used to assign labels to observatons
			by setting one of the probabilities for an observation to 1.0.
			This can be used when only some labels are known by using a
			uniform distribution when the labels are not known. Note that 
			this can be used to assign hard labels, but does not have the 
			same semantics for soft labels, in that it only influences the 
			initial estimate of an observation being generated by a component, 
			not gives a target. Default is None.


		Returns
		-------
		self
		"""

		logp = None
		for i in range(self.max_iter):
			start_time = time.time()

			last_logp = logp
			logp = self.summarize(X, sample_weight=sample_weight, 
				priors=priors)

			if i > 0:
				improvement = logp - last_logp
				duration = time.time() - start_time

				if self.verbose:
					print("[{}] Improvement: {}, Time: {:4.4}s".format(i, 
						improvement, duration))

				if improvement < self.tol:
					break

			self.from_summaries()

		self._reset_cache()
		return self

	def summarize(self, X, sample_weight=None, priors=None):
		"""Extract the sufficient statistics from a batch of data.

		This method calculates the sufficient statistics from optionally
		weighted data and adds them to the stored cache. The examples must be
		given in a 2D format. Sample weights can either be provided as one
		value per example or as a 2D matrix of weights for each feature in
		each example. Labels can be provided for examples but, if provided,
		must be incomplete such that semi-supervised learning can be performed.

		For a mixture model, this step is essentially performing the 'E' part
		of the EM algorithm on a batch of data, where examples are soft-assigned
		to distributions in the model and summaries are derived from that.


		Parameters
		----------
		X: list, tuple, numpy.ndarray, torch.Tensor, shape=(-1, self.d)
			A set of examples to summarize.

		sample_weight: list, tuple, numpy.ndarray, torch.Tensor, optional
			A set of weights for the examples. This can be either of shape
			(-1, self.d) or a vector of shape (-1,). Default is ones.

		priors: list, numpy.ndarray, torch.Tensor, shape=(-1, self.k)
			Prior probabilities of assigning each symbol to each node. If not
			provided, do not include in the calculations (conceptually
			equivalent to a uniform probability, but without scaling the
			probabilities). This can be used to assign labels to observatons
			by setting one of the probabilities for an observation to 1.0.
			Note that this can be used to assign hard labels, but does not
			have the same semantics for soft labels, in that it only
			influences the initial estimate of an observation being generated
			by a component, not gives a target. Default is None.


		Returns
		-------
		logp: float
			The log probability of X given the model.
		"""

		X = _check_parameter(_cast_as_tensor(X), "X", ndim=2)
		if not self._initialized:
			self._initialize(X, sample_weight=sample_weight)

		sample_weight = _reshape_weights(X, _cast_as_tensor(sample_weight, 
			dtype=torch.float32), device=self.device)

		e = self._emission_matrix(X, priors=priors)
		logp = torch.logsumexp(e, dim=1, keepdims=True)
		y = torch.exp(e - logp)

		z = torch.clone(self._w_sum)

		for i, d in enumerate(self.distributions):
			d.summarize(X, y[:, i:i+1] * sample_weight)

			if self.frozen == False:
				self._w_sum[i] = self._w_sum[i] + (y[:, i:i+1] * 
					sample_weight).mean(dim=-1).sum()

		return torch.sum(logp)



# test_factor_graph.py
# Contact: Jacob Schreiber <jmschreiber91@gmail.com>

import numpy
import torch
import pytest

from pomegranate.factor_graph import FactorGraph
from pomegranate.distributions import Exponential
from pomegranate.distributions import Categorical
from pomegranate.distributions import JointCategorical

from .tools import assert_raises
from numpy.testing import assert_array_almost_equal


@pytest.fixture
def X():
	return [[1, 2, 0, 0],
	     [0, 0, 1, 0],
	     [1, 1, 1, 0],
	     [1, 2, 1, 1],
	     [1, 1, 0, 1],
	     [0, 1, 0, 1],
	     [0, 1, 0, 0],
	     [1, 0, 1, 1],
	     [1, 1, 0, 1],
	     [0, 2, 1, 0],
	     [0, 0, 0, 1]]


@pytest.fixture
def w():
	return [0, 1.3, 0, 1.2, 2.5, 6.1, 0, 0, 2.1, 0.3, 3.3]


@pytest.fixture
def X_masked(X):
	mask = torch.tensor(numpy.array([
		[False, True,  True,  True ],
		[True,  True,  False, False],
		[False, False, False, True ],
		[True,  True,  True,  True ],
		[False, True,  False, False],
		[True,  True,  True,  False],
		[False, False, False, True ],
		[True,  False, True,  True ],
		[True,  True,  True,  False],
		[True,  True,  True,  True ],
		[True,  False, True,  False]]))

	X = torch.tensor(numpy.array(X))
	return torch.masked.MaskedTensor(X, mask=mask)


@pytest.fixture
def model():
	f1 = Categorical([[0.23, 0.77]])
	f2 = JointCategorical([
		[[0.10, 0.05, 0.05], [0.15, 0.05, 0.04]],
		[[0.20, 0.10, 0.05], [0.05, 0.10, 0.06]]
	])
	f3 = Categorical([[0.61, 0.39]])
	f4 = JointCategorical([[0.17, 0.15], [0.4, 0.28]])

	m1 = Categorical([[0.5, 0.5]])
	m2 = Categorical([[1./3, 1./3, 1./3]])
	m3 = Categorical([[0.5, 0.5]])
	m4 = Categorical([[0.5, 0.5]])

	model = FactorGraph()
	model.add_factor(f1)
	model.add_factor(f2)
	model.add_factor(f3)
	model.add_factor(f4)

	model.add_marginal(m1)
	model.add_marginal(m2)
	model.add_marginal(m3)
	model.add_marginal(m4)

	model.add_edge(m1, f1)
	model.add_edge(m1, f2)
	model.add_edge(m3, f2)
	model.add_edge(m2, f2)
	model.add_edge(m3, f3)
	model.add_edge(m3, f4)
	model.add_edge(m4, f4)
	return model


@pytest.fixture
def model2():
	f1 = Categorical([[0.23, 0.77]])
	f3 = JointCategorical([[0.17, 0.15], [0.4, 0.28]])
	f4 = JointCategorical([[0.32, 0.12], [0.08, 0.48]])
	f2 = JointCategorical([
		[[0.10, 0.05, 0.05], [0.15, 0.05, 0.04]],
		[[0.20, 0.10, 0.05], [0.05, 0.10, 0.06]]
	])

	m1 = Categorical([[0.5, 0.5]])
	m2 = Categorical([[1./3, 1./3, 1./3]])
	m3 = Categorical([[0.5, 0.5]])
	m4 = Categorical([[0.5, 0.5]])

	model = FactorGraph()
	model.add_factor(f1)
	model.add_factor(f2)
	model.add_factor(f3)
	model.add_factor(f4)

	model.add_marginal(m1)
	model.add_marginal(m2)
	model.add_marginal(m3)
	model.add_marginal(m4)

	model.add_edge(m1, f1)
	model.add_edge(m1, f3)
	model.add_edge(m1, f4)
	model.add_edge(m3, f3)
	model.add_edge(m4, f4)
	model.add_edge(m3, f2)
	model.add_edge(m4, f2)
	model.add_edge(m2, f2)
	return model

###


def test_initialization():
	model = FactorGraph()
	assert len(model.factors) == 0
	assert len(model.marginals) == 0
	assert len(model._factor_idxs) == 0
	assert len(model._marginal_idxs) == 0 

	m = Categorical([[0.5, 0.5]])
	f = Categorical([[0.8, 0.1]])
	model = FactorGraph([f], [m])
	assert len(model.factors) == 1
	assert len(model.marginals) == 1
	assert len(model._factor_idxs) == 1
	assert len(model._marginal_idxs) == 1 
	assert sum(map(len, model._factor_edges)) == 0
	assert sum(map(len, model._marginal_edges)) == 0

	model = FactorGraph([f], [m], [(m, f)])
	assert len(model.factors) == 1
	assert len(model.marginals) == 1
	assert len(model._factor_idxs) == 1
	assert len(model._marginal_idxs) == 1
	assert sum(map(len, model._factor_edges)) == 1
	assert sum(map(len, model._marginal_edges)) == 1


def test_initialization_raises():
	m1 = Categorical([[0.5, 0.5]])
	m2 = Categorical([[0.5, 0.5]])

	f1 = Categorical([[0.8, 0.1]])
	f2 = JointCategorical([[[0.2, 0.1], [0.1, 0.6]]])

	assert_raises(ValueError, FactorGraph, [m1, m2], [f1, f2])
	assert_raises(ValueError, FactorGraph, [f1, f2], [m1, m2], [(f1, f2)])
	assert_raises(ValueError, FactorGraph, [f1, f2], [m1, m2], [(m1, m2)])
	assert_raises(ValueError, FactorGraph, [f1, f2], [m1, m2], [(f1, m1)])
	assert_raises(ValueError, FactorGraph, None, None, [(f1, m1)])


def test_add_factor():
	f1 = Categorical([[0.8, 0.1]])
	f2 = JointCategorical([[[0.2, 0.1], [0.1, 0.6]]])

	model = FactorGraph()
	assert len(model.factors) == 0
	assert model._factor_idxs == {}
	assert model._factor_edges == []

	model.add_factor(f1)
	assert len(model.factors) == 1
	assert model._factor_idxs == {f1: 0}
	assert model._factor_edges == [[]]

	model.add_factor(f2)
	assert len(model.factors) == 2
	assert model._factor_idxs == {f1: 0, f2: 1}
	assert model._factor_edges == [[], []]


def test_add_factor_raises():
	f1 = Categorical([[0.8, 0.1]])
	f2 = JointCategorical([[[0.2, 0.1], [0.1, 0.6]]])

	model = FactorGraph()
	assert_raises(ValueError, model.add_factor, [f1])

	model = FactorGraph()
	assert_raises(ValueError, model.add_factor, None)

	model = FactorGraph()
	assert_raises(ValueError, model.add_factor, Exponential())


def test_add_marginal():
	m1 = Categorical([[0.5, 0.5]])
	m2 = Categorical([[0.5, 0.5]])

	model = FactorGraph()
	assert len(model.marginals) == 0
	assert model._marginal_idxs == {}
	assert model._marginal_edges == []

	model.add_marginal(m1)
	assert len(model.marginals) == 1
	assert model._marginal_idxs == {m1: 0}
	assert model._marginal_edges == [[]]

	model.add_marginal(m2)
	assert len(model.marginals) == 2
	assert model._marginal_idxs == {m1: 0, m2: 1}
	assert model._marginal_edges == [[], []]


def test_add_marginal_raises():
	m = Categorical([[0.5, 0.5]])
	f = JointCategorical([[[0.2, 0.1], [0.1, 0.6]]])

	model = FactorGraph()
	assert_raises(ValueError, model.add_marginal, [m])

	model = FactorGraph()
	assert_raises(ValueError, model.add_marginal, None)

	model = FactorGraph()
	assert_raises(ValueError, model.add_marginal, Exponential())

	model = FactorGraph()
	assert_raises(ValueError, model.add_marginal, f)


def test_add_edge():
	m1 = Categorical([[0.5, 0.5]])
	m2 = Categorical([[0.5, 0.5]])

	f1 = Categorical([[0.8, 0.1]])
	f2 = JointCategorical([[[0.2, 0.1], [0.1, 0.6]]])

	model = FactorGraph([f1, f2], [m1, m2])
	assert model._factor_edges == [[], []]
	assert model._marginal_edges == [[], []]

	model.add_edge(m1, f1)
	assert model._factor_edges == [[0], []]
	assert model._marginal_edges == [[0], []]

	model.add_edge(m1, f2)
	assert model._factor_edges == [[0], [0]]
	assert model._marginal_edges == [[0, 1], []]

	model.add_edge(m2, f2)
	assert model._factor_edges == [[0], [0, 1]]
	assert model._marginal_edges == [[0, 1], [1]]


def test_add_edge_raises():
	m1 = Categorical([[0.5, 0.5]])
	m2 = Categorical([[0.5, 0.5]])
	m3 = Categorical([[0.5, 0.5]])

	f1 = Categorical([[0.8, 0.1]])
	f2 = JointCategorical([[[0.2, 0.1], [0.1, 0.6]]])

	model = FactorGraph([f1, f2], [m1, m2])
	assert_raises(ValueError, model.add_edge, f2, f1)
	assert_raises(ValueError, model.add_edge, f2, m1)
	assert_raises(ValueError, model.add_edge, f1, m1)
	assert_raises(ValueError, model.add_edge, m1, m2)
	assert_raises(ValueError, model.add_edge, m3, f1)
	assert_raises(ValueError, model.add_edge, None, f1)
	assert_raises(ValueError, model.add_edge, m1, None)
	assert_raises(ValueError, model.add_edge, m1, m1)


def test_probability(X, model, model2):
	prob = model.probability(X)
	assert_array_almost_equal(prob, [1.663520e-04, 2.242500e-04, 5.004999e-04, 
		2.102100e-04, 2.935625e-04, 4.384369e-05, 4.968956e-05, 1.751750e-04,
		2.935625e-04, 5.979991e-05, 8.768743e-05])

	prob = model2.probability(X)
	assert_array_almost_equal(prob, [5.133332e-05, 9.199999e-05, 7.186660e-05, 
		2.587201e-04, 3.079999e-04, 9.774998e-06, 2.606666e-05, 2.156001e-04,
		3.079999e-04, 2.299997e-05, 2.932498e-05])


def test_log_probability(X, model, model2):
	prob = model.log_probability(X)
	assert_array_almost_equal(prob, [-8.701405,  -8.402749,  -7.599903,  
		-8.467403,  -8.13342 , -10.03488 ,  -9.909716,  -8.649725,  -8.13342 ,  
		-9.724506, -9.341732])

	prob = model2.log_probability(X)
	assert_array_almost_equal(prob, [-9.877171,  -9.293722,  -9.540699,  
		-8.259764,  -8.085411, -11.535683, -10.554853,  -8.442085,  -8.085411, 
		-10.680017, -10.437071])


def test_predict(X_masked, model, model2):
	y_hat = model.predict(X_masked)
	assert_array_almost_equal(y_hat,
		[[1, 2, 0, 0],
         [0, 0, 1, 0],
         [1, 0, 1, 0],
         [1, 2, 1, 1],
         [1, 1, 1, 0],
         [0, 1, 0, 0],
         [1, 0, 1, 0],
         [1, 1, 1, 1],
         [1, 1, 0, 0],
         [0, 2, 1, 0],
         [0, 0, 0, 0]])

	y_hat = model2.predict(X_masked)
	assert_array_almost_equal(y_hat,
	    [[1, 2, 0, 0],
         [0, 0, 1, 0],
         [1, 0, 1, 0],
         [1, 2, 1, 1],
         [1, 1, 1, 1],
         [0, 1, 0, 0],
         [1, 0, 1, 0],
         [1, 1, 1, 1],
         [1, 1, 0, 1],
         [0, 2, 1, 0],
         [0, 0, 0, 0]])


def test_predict_proba_one_edge():
	m = Categorical([[1./3, 1./3, 1./3]])
	f = Categorical([[0.23, 0.17, 0.60]])

	X = torch.tensor([[0], [0], [1], [2]])
	mask = torch.tensor([[False], [True], [True], [True]])
	X_masked = torch.masked.MaskedTensor(X, mask=mask)

	model = FactorGraph([f], [m], [(m, f)])
	y_hat = model.predict_proba(X_masked)

	assert_array_almost_equal(y_hat[0], 
		[[0.23, 0.17, 0.60],
		 [1.00, 0.00, 0.00],
		 [0.00, 1.00, 0.00],
		 [0.00, 0.00, 1.00]])


def test_predict_proba_one_edge_raises():
	m = Categorical([[1./3, 1./3, 1./3]])
	f = Categorical([[0.23, 0.17, 0.60]])
	model = FactorGraph([f], [m], [(m, f)])

	X = torch.tensor([[0, 0]])
	mask = torch.tensor([[False, True]])
	X_masked = torch.masked.MaskedTensor(X, mask=mask)
	assert_raises(ValueError, model.predict_proba, X_masked)

	X = torch.tensor([[3]])
	mask = torch.tensor([[True]])
	X_masked = torch.masked.MaskedTensor(X, mask=mask)
	assert_raises(IndexError, model.predict_proba, X_masked)

	mask = torch.tensor([[False]])
	X_masked = torch.masked.MaskedTensor(X, mask=mask)
	model.predict_proba(X_masked)


def test_predict_proba_one_edge_nonuniform():
	m = Categorical([[0.40, 0.11, 0.49]])
	f = Categorical([[0.23, 0.17, 0.60]])

	X = torch.tensor([[0], [0]])
	mask = torch.tensor([[False], [True]])
	X_masked = torch.masked.MaskedTensor(X, mask=mask)

	model = FactorGraph([f], [m], [(m, f)])
	y_hat = model.predict_proba(X_masked)

	z = 0.23*0.40 + 0.17*0.11 + 0.60*0.49

	assert_array_almost_equal(y_hat[0], 
		[[0.23*0.40/z, 0.17*0.11/z, 0.60*0.49/z],
		 [1.00, 0.00, 0.00]])


def test_predict_proba_monty_hall():
	m1 = Categorical([[1./3, 1./3, 1./3]])
	m2 = Categorical([[1./3, 1./3, 1./3]])
	m3 = Categorical([[1./3, 1./3, 1./3]])

	p = numpy.array([
		 [[0.0, 0.5, 0.5], [0.0, 0.0, 1.0], [0.0, 1.0, 0.0]], 
		 [[0.0, 0.0, 1.0], [0.5, 0.0, 0.5], [1.0, 0.0, 0.0]],
		 [[0.0, 1.0, 0.0], [1.0, 0.0, 0.0], [0.5, 0.5, 0.0]]
	])


	f1 = Categorical([[1./3, 1./3, 1./3]])
	f2 = Categorical([[1./3, 1./3, 1./3]])
	f3 = JointCategorical(p / 9.) 

	X = torch.tensor([
		[0, 0, 0],
		[0, 1, 0],
		[0, 2, 0],
		[1, 0, 0],
		[2, 0, 0]
	])

	mask = torch.tensor([
		[False, False, False],
		[False, True,  False],
		[False, True,  False],
		[True,  False, False],
		[True,  False, False]
	])

	X_masked = torch.masked.MaskedTensor(X, mask=mask)

	model = FactorGraph([f1, f2, f3], [m1, m2, m3], [(m1, f1), (m1, f3), 
		(m2, f3), (m3, f3), (m2, f2)])
	y_hat = model.predict_proba(X_masked)

	assert_array_almost_equal(y_hat[0], 
		[[1./3, 1./3, 1./3],
		 [1./3, 1./3, 1./3],
		 [1./3, 1./3, 1./3],
		 [0.0, 1.0, 0.0],
		 [0.0, 0.0, 1.0]])

	assert_array_almost_equal(y_hat[1],
		[[1./3, 1./3, 1./3],
		 [0.0, 1.0, 0.0],
		 [0.0, 0.0, 1.0],
		 [1./3, 1./3, 1./3],
		 [1./3, 1./3, 1./3]])

	assert_array_almost_equal(y_hat[2],
		[[1./3, 1./3, 1./3],
         [0.5000, 0.0000, 0.5000],
         [0.5000, 0.5000, 0.0000],
         [0.5000, 0.0000, 0.5000],
         [0.5000, 0.5000, 0.0000]])


	X = torch.tensor([
		[0, 0, 0],
		[0, 1, 0],
		[0, 2, 0],
		[1, 0, 0],
		[2, 0, 0],
		[0, 0, 1],
		[0, 1, 0],
		[2, 1, 0]
	])

	mask = torch.tensor([
		[True,  True,  False],
		[True,  True,  False],
		[True,  True,  False],
		[True,  True, False],
		[True,  True, False],
		[False, True,  True],
		[False, False, True],
		[True , False, True]
	])

	X_masked = torch.masked.MaskedTensor(X, mask=mask)
	y_hat = model.predict_proba(X_masked)

	assert_array_almost_equal(y_hat[0], 
		[[1.0000, 0.0000, 0.0000],
         [1.0000, 0.0000, 0.0000],
         [1.0000, 0.0000, 0.0000],
         [0.0000, 1.0000, 0.0000],
         [0.0000, 0.0000, 1.0000],
         [1./3  , 0.0000, 2./3  ],
         [0.0000, 0.5000, 0.5000],
         [0.0000, 0.0000, 1.0000]])

	assert_array_almost_equal(y_hat[1],
		[[1.0000, 0.0000, 0.0000],
         [0.0000, 1.0000, 0.0000],
         [0.0000, 0.0000, 1.0000],
         [1.0000, 0.0000, 0.0000],
         [1.0000, 0.0000, 0.0000],
         [1.0000, 0.0000, 0.0000],
         [0.0000, 0.5000, 0.5000],
         [0.0000, 2./3  , 1./3  ]])

	assert_array_almost_equal(y_hat[2],
		[[0.0000, 0.5000, 0.5000],
         [0.0000, 0.0000, 1.0000],
         [0.0000, 1.0000, 0.0000],
         [0.0000, 0.0000, 1.0000],
         [0.0000, 1.0000, 0.0000],
         [0.0000, 1.0000, 0.0000],
         [1.0000, 0.0000, 0.0000],
         [1.0000, 0.0000, 0.0000]])


def test_predict_proba_simple(model, X_masked):
	y_hat = model.predict_proba(X_masked)

	assert_array_almost_equal(y_hat[0], 
		[[0.2300, 0.7700],
         [1.0000, 0.0000],
         [0.2011, 0.7989],
         [0.0000, 1.0000],
         [0.1299, 0.8701],
         [1.0000, 0.0000],
         [0.2011, 0.7989],
         [0.0000, 1.0000],
         [0.0000, 1.0000],
         [1.0000, 0.0000],
         [1.0000, 0.0000]], 4)

	assert_array_almost_equal(y_hat[1],
		[[0.0000, 0.0000, 1.0000],
         [1.0000, 0.0000, 0.0000],
         [0.4469, 0.3453, 0.2078],
         [0.0000, 0.0000, 1.0000],
         [0.0000, 1.0000, 0.0000],
         [0.0000, 1.0000, 0.0000],
         [0.4469, 0.3453, 0.2078],
         [0.2381, 0.4762, 0.2857],
         [0.0000, 1.0000, 0.0000],
         [0.0000, 0.0000, 1.0000],
         [0.5000, 0.2500, 0.2500]], 4)

	assert_array_almost_equal(y_hat[2],
		[[1.0000, 0.0000],
         [0.3292, 0.6708],
         [0.4916, 0.5084],
         [0.0000, 1.0000],
         [0.4240, 0.5760],
         [1.0000, 0.0000],
         [0.4916, 0.5084],
         [0.0000, 1.0000],
         [1.0000, 0.0000],
         [0.0000, 1.0000],
         [1.0000, 0.0000]], 4)

	assert_array_almost_equal(y_hat[3],
		[[1.0000, 0.0000],
         [0.5695, 0.4305],
         [1.0000, 0.0000],
         [0.0000, 1.0000],
         [0.5641, 0.4359],
         [0.5312, 0.4688],
         [1.0000, 0.0000],
         [0.0000, 1.0000],
         [0.5312, 0.4688],
         [1.0000, 0.0000],
         [0.5312, 0.4688]], 4)


def test_predict_proba_cycle(model2, X_masked):
	y_hat = model2.predict_proba(X_masked)

	assert_array_almost_equal(y_hat[0], 
		[[0.3368, 0.6632],
         [1.0000, 0.0000],
         [0.3673, 0.6327],
         [0.0000, 1.0000],
         [0.1031, 0.8969],
         [1.0000, 0.0000],
         [0.3673, 0.6327],
         [0.0000, 1.0000],
         [0.0000, 1.0000],
         [1.0000, 0.0000],
         [1.0000, 0.0000]], 4)

	assert_array_almost_equal(y_hat[1],
		[[0.0000, 0.0000, 1.0000],
         [1.0000, 0.0000, 0.0000],
         [0.5408, 0.2704, 0.1888],
         [0.0000, 0.0000, 1.0000],
         [0.0000, 1.0000, 0.0000],
         [0.0000, 1.0000, 0.0000],
         [0.5408, 0.2704, 0.1888],
         [0.2381, 0.4762, 0.2857],
         [0.0000, 1.0000, 0.0000],
         [0.0000, 0.0000, 1.0000],
         [0.5388, 0.2371, 0.2241]], 4)

	assert_array_almost_equal(y_hat[2],
		[[1.0000, 0.0000],
         [0.4474, 0.5526],
         [0.4287, 0.5713],
         [0.0000, 1.0000],
         [0.4110, 0.5890],
         [1.0000, 0.0000],
         [0.4287, 0.5713],
         [0.0000, 1.0000],
         [1.0000, 0.0000],
         [0.0000, 1.0000],
         [1.0000, 0.0000]], 4)

	assert_array_almost_equal(y_hat[3],
		[[1.0000, 0.0000],
         [0.7916, 0.2084],
         [1.0000, 0.0000],
         [0.0000, 1.0000],
         [0.2031, 0.7969],
         [0.7273, 0.2727],
         [1.0000, 0.0000],
         [0.0000, 1.0000],
         [0.1429, 0.8571],
         [1.0000, 0.0000],
         [0.6897, 0.3103]], 4)


def test_fit(X, model, model2):
	model.fit(X)
	assert_array_almost_equal(model.factors[0].probs, [[0.4545, 0.5455]], 4)
	assert_array_almost_equal(model.factors[1].probs, 
		[[[0.0909, 0.1818, 0.0000],
          [0.0909, 0.0000, 0.0909]],

         [[0.0000, 0.1818, 0.0909],
          [0.0909, 0.0909, 0.0909]]], 4)
	assert_array_almost_equal(model.factors[2].probs, [[0.5455, 0.4545]], 4)
	assert_array_almost_equal(model.factors[3].probs, 
		[[0.1818, 0.3636],
         [0.2727, 0.1818]], 4)

	assert_array_almost_equal(model.marginals[0].probs, [[0.5, 0.5]])
	assert_array_almost_equal(model.marginals[1].probs, [[1./3, 1./3, 1./3]])
	assert_array_almost_equal(model.marginals[2].probs, [[0.5, 0.5]])
	assert_array_almost_equal(model.marginals[3].probs, [[0.5, 0.5]])


	model2.fit(X)
	assert_array_almost_equal(model2.factors[0].probs, [[0.4545, 0.5455]], 4)
	assert_array_almost_equal(model2.factors[1].probs, 
		[[[0.0000, 0.0909, 0.0909],
          [0.0909, 0.2727, 0.0000]],

         [[0.0909, 0.0909, 0.0909],
          [0.0909, 0.0000, 0.0909]]], 4)
	assert_array_almost_equal(model2.factors[2].probs, 
		[[0.2727, 0.1818],
         [0.2727, 0.2727]], 4)
	assert_array_almost_equal(model2.factors[3].probs, 
		[[0.2727, 0.1818],
         [0.1818, 0.3636]], 4)

	assert_array_almost_equal(model2.marginals[0].probs, [[0.5, 0.5]])
	assert_array_almost_equal(model2.marginals[1].probs, [[1./3, 1./3, 1./3]])
	assert_array_almost_equal(model2.marginals[2].probs, [[0.5, 0.5]])
	assert_array_almost_equal(model2.marginals[3].probs, [[0.5, 0.5]])


def test_fit_weighted(X, w, model, model2):
	model.fit(X, sample_weight=w)

	assert_array_almost_equal(model.factors[0].probs, [[0.6548, 0.3452]], 4)
	assert_array_almost_equal(model.factors[1].probs, 
		[[[0.1964, 0.3631, 0.0000],
          [0.0774, 0.0000, 0.0179]],

         [[0.0000, 0.2738, 0.0000],
          [0.0000, 0.0000, 0.0714]]], 4)
	assert_array_almost_equal(model.factors[2].probs, [[0.8333, 0.1667]], 4)
	assert_array_almost_equal(model.factors[3].probs, 
		[[0.0000, 0.8333],
         [0.0952, 0.0714]], 4)

	assert_array_almost_equal(model.marginals[0].probs, [[0.5, 0.5]])
	assert_array_almost_equal(model.marginals[1].probs, [[1./3, 1./3, 1./3]])
	assert_array_almost_equal(model.marginals[2].probs, [[0.5, 0.5]])
	assert_array_almost_equal(model.marginals[3].probs, [[0.5, 0.5]])


	model2.fit(X, sample_weight=w)
	assert_array_almost_equal(model2.factors[0].probs, [[0.6548, 0.3452]], 4)
	assert_array_almost_equal(model2.factors[1].probs, 
		[[[0.0000, 0.0000, 0.0000],
          [0.1964, 0.6369, 0.0000]],

         [[0.0774, 0.0000, 0.0179],
          [0.0000, 0.0000, 0.0714]]], 4)
	assert_array_almost_equal(model2.factors[2].probs, 
		[[0.5595, 0.0952],
         [0.2738, 0.0714]], 4)
	assert_array_almost_equal(model2.factors[3].probs, 
		[[0.0952, 0.5595],
         [0.0000, 0.3452]], 4)

	assert_array_almost_equal(model2.marginals[0].probs, [[0.5, 0.5]])
	assert_array_almost_equal(model2.marginals[1].probs, [[1./3, 1./3, 1./3]])
	assert_array_almost_equal(model2.marginals[2].probs, [[0.5, 0.5]])
	assert_array_almost_equal(model2.marginals[3].probs, [[0.5, 0.5]])


def test_summarize(X, model, model2):
	model.summarize(X)
	assert_array_almost_equal(model.factors[0].probs, [[0.23, 0.77]])
	assert_array_almost_equal(model.factors[1].probs, [
		[[0.10, 0.05, 0.05], [0.15, 0.05, 0.04]],
		[[0.20, 0.10, 0.05], [0.05, 0.10, 0.06]]
	])
	assert_array_almost_equal(model.factors[2].probs, [[0.61, 0.39]])
	assert_array_almost_equal(model.factors[3].probs, 
		[[0.17, 0.15], [0.4, 0.28]])

	assert_array_almost_equal(model.marginals[0].probs, [[0.5, 0.5]])
	assert_array_almost_equal(model.marginals[1].probs, [[1./3, 1./3, 1./3]])
	assert_array_almost_equal(model.marginals[2].probs, [[0.5, 0.5]])
	assert_array_almost_equal(model.marginals[3].probs, [[0.5, 0.5]])


	model2.summarize(X)
	assert_array_almost_equal(model2.factors[0].probs, [[0.23, 0.77]], 4)
	assert_array_almost_equal(model2.factors[1].probs, [
		[[0.10, 0.05, 0.05], [0.15, 0.05, 0.04]],
		[[0.20, 0.10, 0.05], [0.05, 0.10, 0.06]]
	])
	assert_array_almost_equal(model2.factors[2].probs, 
		[[0.17, 0.15], [0.4, 0.28]])
	assert_array_almost_equal(model2.factors[3].probs, 
		[[0.32, 0.12], [0.08, 0.48]])

	assert_array_almost_equal(model2.marginals[0].probs, [[0.5, 0.5]])
	assert_array_almost_equal(model2.marginals[1].probs, [[1./3, 1./3, 1./3]])
	assert_array_almost_equal(model2.marginals[2].probs, [[0.5, 0.5]])
	assert_array_almost_equal(model2.marginals[3].probs, [[0.5, 0.5]])


def test_from_summarize(X, model, model2):
	model.summarize(X)
	model.from_summaries()
	assert_array_almost_equal(model.factors[0].probs, [[0.4545, 0.5455]], 4)
	assert_array_almost_equal(model.factors[1].probs, 
		[[[0.0909, 0.1818, 0.0000],
          [0.0909, 0.0000, 0.0909]],

         [[0.0000, 0.1818, 0.0909],
          [0.0909, 0.0909, 0.0909]]], 4)
	assert_array_almost_equal(model.factors[2].probs, [[0.5455, 0.4545]], 4)
	assert_array_almost_equal(model.factors[3].probs, 
		[[0.1818, 0.3636],
         [0.2727, 0.1818]], 4)

	assert_array_almost_equal(model.marginals[0].probs, [[0.5, 0.5]])
	assert_array_almost_equal(model.marginals[1].probs, [[1./3, 1./3, 1./3]])
	assert_array_almost_equal(model.marginals[2].probs, [[0.5, 0.5]])
	assert_array_almost_equal(model.marginals[3].probs, [[0.5, 0.5]])


	model2.summarize(X)
	model2.from_summaries()
	assert_array_almost_equal(model2.factors[0].probs, [[0.4545, 0.5455]], 4)
	assert_array_almost_equal(model2.factors[1].probs, 
		[[[0.0000, 0.0909, 0.0909],
          [0.0909, 0.2727, 0.0000]],

         [[0.0909, 0.0909, 0.0909],
          [0.0909, 0.0000, 0.0909]]], 4)
	assert_array_almost_equal(model2.factors[2].probs, 
		[[0.2727, 0.1818],
         [0.2727, 0.2727]], 4)
	assert_array_almost_equal(model2.factors[3].probs, 
		[[0.2727, 0.1818],
         [0.1818, 0.3636]], 4)

	assert_array_almost_equal(model2.marginals[0].probs, [[0.5, 0.5]])
	assert_array_almost_equal(model2.marginals[1].probs, [[1./3, 1./3, 1./3]])
	assert_array_almost_equal(model2.marginals[2].probs, [[0.5, 0.5]])
	assert_array_almost_equal(model2.marginals[3].probs, [[0.5, 0.5]])



# factor_graph.py
# Author: Jacob Schreiber <jmschreiber91@gmail.com>

import time
import torch
import itertools

from ._utils import _cast_as_tensor
from ._utils import _update_parameter
from ._utils import _check_parameter
from ._utils import _reshape_weights

from .distributions._distribution import Distribution
from .distributions._distribution import ConditionalDistribution

from .distributions import Categorical
from .distributions import JointCategorical


class FactorGraph(Distribution):
	"""A factor graph object.

	A factor graph represents a probability distribution as a bipartite graph
	where marginal distributions of each dimension in the distribution are on
	one side of the graph and factors are on the other side. The distributions
	on the factor side encode probability estimates from the model, whereas the
	distributions on the marginal side encode probability estimates from the
	data. 

	Inference is done on the factor graph using the loopy belief propagation
	algorithm. This is an iterative algorithm where "messages" are passed
	along each edge between the marginals and the factors until the estimates
	for the marginals converges. In brief: each message represents what the
	generating node thinks its marginal distribution is with respect to the
	child node. Calculating each message involves marginalizing the parent
	node with respect to *every other* node. When the parent node is already
	a univariate distribution -- either because it is a marginal node or
	a univariate factor node -- no marginalization is needed and it sends
	itself as the message. Basically, a joint probability table will receive
	messages from all the marginal nodes that comprise its dimensions and,
	to each of those marginal nodes, it will send a message back saying what
	it (the joint probability table) thinks its marginal distribution is with
	respect to the messages from the OTHER marginals. More concretely, if a
	joint probability table has two dimensions with marginal node parents
	A and B, it will send a message to A that is itself after marginalizing out
	B, and will send a message to B that is itself after marginalizing out A. 

	..note:: It is worth noting that this algorithm is exact when the structure
	is a tree. If there exist any loops in the factors, i.e., you can draw a
	circle beginning with a factor and then hopping between marginals and
	factors and make it back to the factor without crossing any edges twice,
	the probabilities returned are approximate.


	Parameters
	----------
	factors: tuple or list or None
		A set of distribution objects. These do not need to be initialized,
		i.e. can be "Categorical()". Currently, they must be either Categorical
		or JointCategorical distributions. Default is None.

	marginals: tuple or list or None
		A set of distribution objects. These must be initialized and be
		Categorical distributions.

	edges: list or tuple or None
		A set of edges. Critically, the items in this list must be the
		distribution objects themselves, and the order that edges must match
		the order distributions in a multivariate distribution. Specifically,
		if you have a multivariate distribution, the first edge that includes
		it must correspond to the first dimension, the second edge must
		correspond to the second dimension, etc, and the total number of
		edges cannot exceed the number of dimensions. Default is None.

	max_iter: int, optional
		The number of iterations to do in the inference step as distributions
		are converging. Default is 10.

	tol: float, optional
		The threshold at which to stop during fitting when the improvement
		goes under. Default is 1e-6.

	inertia: float, [0, 1], optional
		Indicates the proportion of the update to apply to the parameters
		during training. When the inertia is 0.0, the update is applied in
		its entirety and the previous parameters are ignored. When the
		inertia is 1.0, the update is entirely ignored and the previous
		parameters are kept, equivalently to if the parameters were frozen.

	frozen: bool, optional
		Whether all the parameters associated with this distribution are frozen.
		If you want to freeze individual pameters, or individual values in those
		parameters, you must modify the `frozen` attribute of the tensor or
		parameter directly. Default is False.

	check_data: bool, optional
		Whether to check properties of the data and potentially recast it to
		torch.tensors. This does not prevent checking of parameters but can
		slightly speed up computation when you know that your inputs are valid.
		Setting this to False is also necessary for compiling.

	verbose: bool, optional
		Whether to print the improvement and timings during training.
	"""

	def __init__(self, factors=None, marginals=None, edges=None, max_iter=20, 
		tol=1e-6, inertia=0.0, frozen=False, check_data=True, verbose=False):
		super().__init__(inertia=inertia, frozen=frozen, check_data=check_data)
		self.name = "FactorGraph"

		self.factors = torch.nn.ModuleList([])
		self.marginals = torch.nn.ModuleList([])

		self._factor_idxs = {}
		self._marginal_idxs = {}

		self._factor_edges = []
		self._marginal_edges = []

		self.max_iter = _check_parameter(max_iter, "max_iter", min_value=1, 
			dtypes=[int, torch.int16, torch.int32, torch.int64])
		self.tol = _check_parameter(tol, "tol", min_value=0)
		self.verbose = verbose

		self.d = 0
		self._initialized = factors is not None and factors[0]._initialized

		if factors is not None:
			_check_parameter(factors, "factors", dtypes=(list, tuple))

			for factor in factors:
				self.add_factor(factor)

		if marginals is not None:
			_check_parameter(marginals, "marginals", dtypes=(list, tuple))

			for marginal in marginals:
				self.add_marginal(marginal)

		if edges is not None:
			_check_parameter(edges, "edges", dtypes=(list, tuple))

			for marginal, factor in edges:
				self.add_edge(marginal, factor)

		self._initialized = not factors

	def _initialize(self, d):
		self._initialized = True
		super()._initialize(d)

	def _reset_cache(self):
		return

	def add_factor(self, distribution):
		"""Adds a distribution to the set of factors.

		
		Parameters
		----------
		distribution: pomegranate.distributions.Distribution
			A distribution object to include as a node.
		"""

		if not isinstance(distribution, (Categorical, JointCategorical)):
			raise ValueError("Must be a Categorical or a JointCategorical"
				" distribution.")

		self.factors.append(distribution)
		self._factor_edges.append([])
		self._factor_idxs[distribution] = len(self.factors) - 1

		self._initialized = distribution._initialized

	def add_marginal(self, distribution):
		"""Adds a distribution to the set of marginals.

		This adds a distribution to the marginal side of the bipartate graph.
		This distribution must be univariate. 

		Parameters
		----------
		distribution: pomegranate.distributions.Distribution
			A distribution object to include as a node.
		"""

		if not isinstance(distribution, Categorical):
			raise ValueError("Must be a Categorical distribution.")
		
		self.marginals.append(distribution)
		self._marginal_edges.append([])
		self._marginal_idxs[distribution] = len(self.marginals) - 1

		self.d += 1

	def add_edge(self, marginal, factor):
		"""Adds an undirected edge to the set of edges.

		Because a factor graph is a bipartite graph, one of the edges must be
		a marginal distribution and the other edge must be a factor 
		distribution.

		Parameters
		----------
		marginal: pomegranate.distributions.Distribution
			The marginal distribution to include in the edge.

		factor: pomegranate.distributions.Distribution
			The factor distribution to include in the edge.
		"""

		if marginal not in self._marginal_idxs:
			raise ValueError("Marginal distribution does not exist in graph.")
		
		if factor not in self._factor_idxs:
			raise ValueError("Factor distribution does not exist in graph.")

		m_idx = self._marginal_idxs[marginal]
		f_idx = self._factor_idxs[factor]

		self._factor_edges[f_idx].append(m_idx)
		self._marginal_edges[m_idx].append(f_idx)

	def log_probability(self, X):
		"""Calculate the log probability of each example.

		This method calculates the log probability of each example given the
		parameters of the distribution. The examples must be given in a 2D
		format.


		Parameters
		----------
		X: list, tuple, numpy.ndarray, torch.Tensor, shape=(-1, self.d)
			A set of examples to evaluate.

		Returns
		-------
		logp: torch.Tensor, shape=(-1,)
			The log probability of each example.
		"""

		X = _check_parameter(_cast_as_tensor(X), "X", ndim=2, 
			check_parameter=self.check_data)
		
		logps = torch.zeros(X.shape[0], device=X.device, dtype=torch.float32)
		for idxs, factor in zip(self._factor_edges, self.factors):
			logps += factor.log_probability(X[:, idxs])

		for i, marginal in enumerate(self.marginals):
			logps += marginal.log_probability(X[:, i:i+1])

		return logps

	def predict(self, X):
		"""Infers the maximum likelihood value for each missing value.

		This method infers a probability distribution for each of the missing
		values in the data. First, the sum-product algorithm is run to infer
		a probability distribution for each variable. Then, the maximum
		likelihood value is returned from that distribution.

		The input to this method must be a torch.masked.MaskedTensor where the
		mask specifies which variables are observed (mask = True) and which ones
		are not observed (mask = False) for each of the values. When setting
		mask = False, it does not matter what the corresponding value in the
		tensor is. Different sets of variables can be observed or missing in
		different examples. 

		Unlike the `predict_proba` and `predict_log_proba` methods, this
		method preserves the dimensions of the original data because it does
		not matter how many categories a variable can take when you're only
		returning the maximally likely one.


		Parameters
		----------
		X: torch.masked.MaskedTensor
			A masked tensor where the observed values are available and the
			unobserved values are missing, i.e., the mask is True for
			observed values and the mask is False for missing values. It does
			not matter what the underlying value in the tensor is for the 
			missing values.
		"""

		y = [t.argmax(dim=1) for t in self.predict_proba(X)]
		return torch.vstack(y).T.contiguous()

	def predict_proba(self, X):
		"""Predict the probability of each variable given some evidence.

		Given some evidence about the value that each variable takes, infer
		the value that each other variable takes. If no evidence is given,
		this returns the marginal value of each variable given the dependence
		structure in the network.

		Currently, only hard evidence is supported, where the evidence takes
		the form of a discrete value. The evidence is represented as a
		masked tensor where the masked out values are considered missing.


		Parameters
		----------
		X: torch.masked.MaskedTensor
			A masked tensor where the observed values are available and the
			unobserved values are missing, i.e., the mask is True for
			observed values and the mask is False for missing values. It does
			not matter what the underlying value in the tensor is for the 
			missing values.
		"""

		nm = len(self.marginals)
		nf = len(self.factors)

		if X.shape[1] != nm:
			raise ValueError("X.shape[1] must match the number of marginals.")

		# Pull out the current distributions from factors and marginals
		factors = []
		marginals = []
		prior_marginals = []
		current_marginals = []

		for i, m in enumerate(self.marginals):
			p = torch.clone(m.probs[0])
			p = p.repeat((X.shape[0],) + tuple(1 for _ in p.shape))

			# Use the evidence in the data to set marginal distributions
			for j in range(X.shape[0]):
				if X._masked_mask[j, i] == True:
					value = X._masked_data[j, i]

					p[j] = 0
					p[j, value] = 1.0

			marginals.append(p)
			prior_marginals.append(torch.clone(p))
			current_marginals.append(torch.clone(p))

		for i, f in enumerate(self.factors):
			if not isinstance(f, Categorical):
				p = torch.clone(f.probs)
			else:
				p = torch.clone(f.probs[0])

			p = p.repeat((X.shape[0],) + tuple(1 for _ in p.shape))
			factors.append(p)

		# Set the original in and out messages along the edges
		in_messages, out_messages = [], []

		for i, m in enumerate(marginals):
			k = len(self._marginal_edges[i])
			in_messages.append([])

			for j in range(k):
				in_messages[-1].append(m)

		for i in range(len(factors)):
			k = len(self._factor_edges[i])
			out_messages.append([])
			
			for j in range(k):
				marginal_idx = self._factor_edges[i][j]
				d_j = marginals[marginal_idx]
				out_messages[-1].append(d_j)


		# Begin iterations
		iteration = 0
		while iteration < self.max_iter:
			# Update the messages going into the nodes.
			for i, f in enumerate(factors):
				ni_edges = len(self._factor_edges[i])

				for k in range(ni_edges):
					message = torch.clone(f)

					shape = torch.ones(len(message.shape), dtype=torch.int32)
					shape[0] = X.shape[0]

					for l in range(ni_edges):
						if k == l:
							continue

						shape[l+1] = message.shape[l+1]						
						message *= out_messages[i][l].reshape(*shape)
						message = torch.sum(message, dim=l+1, keepdims=True)
						shape[l+1] = 1

					else:
						message = message.squeeze()
						if len(message.shape) == 1:
							message = message.unsqueeze(0)

					j = self._factor_edges[i][k]
					for ik, parent in enumerate(self._marginal_edges[j]):
						if parent == i:
							dims = tuple(range(1, len(message.shape)))
							in_messages[j][ik] = message / message.sum(
								dim=dims, keepdims=True)

							break


			# Calculate the current estimates of the marginals
			loss = 0
			for i, m in enumerate(marginals):
				current_marginals[i] = torch.clone(m)

				for k in range(len(self._marginal_edges[i])):
					current_marginals[i] *= in_messages[i][k]


				dims = tuple(range(1, len(current_marginals[i].shape)))
				current_marginals[i] /= current_marginals[i].sum(dim=dims, 
					keepdims=True)

				loss += torch.nn.KLDivLoss(reduction="batchmean")(torch.log(
					current_marginals[i] + 1e-8), prior_marginals[i])

			if self.verbose:
				print(iteration, loss.item())

			if loss < self.tol:
				break


			# Update the messages leaving based on the new marginals
			for i, m in enumerate(marginals):
				ni_edges = len(self._marginal_edges[i])

				for k in range(ni_edges):
					message = torch.clone(m)

					for l in range(ni_edges):
						if k == l:
							continue

						message *= in_messages[i][l]

					j = self._marginal_edges[i][k]
					for ik, parent in enumerate(self._factor_edges[j]):
						if parent == i:
							dims = tuple(range(1, len(message.shape)))
							out_messages[j][ik] = message / message.sum(
								dim=dims, keepdims=True)
							break

			prior_marginals = [torch.clone(d) for d in current_marginals]
			iteration += 1


		return current_marginals

	def predict_log_proba(self, X):
		"""Infers the probability of each category given the model and data.

		This method is a wrapper around the `predict_proba` method and simply
		takes the log of each returned tensor.

		This method infers a log probability distribution for each of the 
		missing  values in the data. It uses the factor graph representation of 
		the Bayesian network to run the sum-product/loopy belief propagation
		algorithm.

		The input to this method must be a torch.masked.MaskedTensor where the
		mask specifies which variables are observed (mask = True) and which ones
		are not observed (mask = False) for each of the values. When setting
		mask = False, it does not matter what the corresponding value in the
		tensor is. Different sets of variables can be observed or missing in
		different examples. 

		An important note is that, because each variable can have a different
		number of categories in the categorical setting, the return is a list
		of tensors where each element in that list is the marginal probability
		distribution for that variable. More concretely: the first element will
		be the distribution of values for the first variable across all
		examples. When the first variable has been provided as evidence, the
		distribution will be clamped to the value provided as evidence.

		..warning:: This inference is exact given a Bayesian network that has
		a tree-like structure, but is only approximate for other cases. When
		the network is acyclic, this procedure will converge, but if the graph
		contains cycles then there is no guarantee on convergence.


		Parameters
		----------
		X: torch.masked.MaskedTensor, shape=(-1, d)
			The data to predict values for. The mask should correspond to
			whether the variable is observed in the example. 
		

		Returns
		-------
		y: list of tensors, shape=(d,)
			A list of tensors where each tensor contains the distribution of
			values for that dimension.
		"""

		return [torch.log(t) for t in self.predict_proba(X)]

	def fit(self, X, sample_weight=None):
		"""Fit the factors of the model to optionally weighted examples.

		This method will fit the provided factor distributions to the given
		data and their optional weights. It will not update the marginal
		distributions, as that information is already encoded in the joint
		probabilities.

		..note:: A structure must already be provided. Currently, structure
		learning of factor graphs is not supported.


		Parameters
		----------
		X: list, tuple, numpy.ndarray, torch.Tensor, shape=(-1, self.d)
			A set of examples to evaluate. 

		sample_weight: list, tuple, numpy.ndarray, torch.Tensor, optional
			A set of weights for the examples. This can be either of shape
			(-1, self.d) or a vector of shape (-1,). Default is ones.


		Returns
		-------
		self
		"""

		self.summarize(X, sample_weight=sample_weight)
		self.from_summaries()
		return self

	def summarize(self, X, sample_weight=None):
		"""Extract the sufficient statistics from a batch of data.

		This method calculates the sufficient statistics from optionally
		weighted data and adds them to the stored cache for each distribution
		in the network. Sample weights can either be provided as one
		value per example or as a 2D matrix of weights for each feature in
		each example.


		Parameters
		----------
		X: list, tuple, numpy.ndarray, torch.Tensor, shape=(-1, len, self.d)
			A set of examples to summarize.

		sample_weight: list, tuple, numpy.ndarray, torch.Tensor, optional
			A set of weights for the examples. This can be either of shape
			(-1, self.d) or a vector of shape (-1,). Default is ones.


		Returns
		-------
		logp: torch.Tensor, shape=(-1,)
			The log probability of each example.
		"""

		if self.frozen:
			return

		X, sample_weight = super().summarize(X, sample_weight=sample_weight)
		X = _check_parameter(X, "X", ndim=2, check_parameter=self.check_data)

		for i, factor in enumerate(self.factors):
			factor.summarize(X[:, self._factor_edges[i]], 
				sample_weight=sample_weight[:,i])

	def from_summaries(self):
		if self.frozen:
			return

		for distribution in self.factors:
			distribution.from_summaries()



# categorical.py
# Contact: Jacob Schreiber <jmschreiber91@gmail.com>

import torch

from .._utils import _inplace_add
from .._utils import _cast_as_tensor
from .._utils import _cast_as_parameter
from .._utils import _update_parameter
from .._utils import _check_parameter
from .._utils import _reshape_weights

from ._distribution import Distribution


class Categorical(Distribution):
	"""A categorical distribution object.

	A categorical distribution models the probability of a set of distinct
	values happening. It is an extension of the Bernoulli distribution to
	multiple values. Sometimes it is referred to as a discrete distribution,
	but this distribution does not enforce that the numeric values used for the
	keys have any relationship based on their identity. Permuting the keys will
	have no effect on the calculation. This distribution assumes that the
	features are independent from each other.

	The keys must be contiguous non-negative integers that begin at zero. 
	Because the probabilities are represented as a single tensor, each feature
	must have values for all keys up to the maximum key of any one distribution.
	Specifically, if one feature has 10 keys and a second feature has only 4,
	the tensor must go out to 10 for each feature but encode probabilities of
	zero for the second feature. 


	Parameters
	----------
	probs: list, numpy.ndarray, torch.tensor or None, shape=(k, d), optional
		Probabilities for each key for each feature, where k is the largest
		number of keys across all features. Default is None

	n_categories: list, numpy.ndarray, torch.tensor or None, optional
		The number of categories for each feature in the data. Only needs to
		be provided when the parameters will be learned directly from data and
		you want to make sure that right number of keys are included in each
		dimension. Default is None.

	pseudocount: float, optional
		A value to add to the observed counts of each feature when training.
		Setting this to a positive value ensures that no probabilities are
		truly zero. Default is 0.

	inertia: float, (0, 1), optional
		Indicates the proportion of the update to apply to the parameters
		during training. When the inertia is 0.0, the update is applied in
		its entirety and the previous parameters are ignored. When the
		inertia is 1.0, the update is entirely ignored and the previous
		parameters are kept, equivalently to if the parameters were frozen.

	frozen: bool, optional
		Whether all the parameters associated with this distribution are frozen.
		If you want to freeze individual pameters, or individual values in those
		parameters, you must modify the `frozen` attribute of the tensor or
		parameter directly. Default is False.

	check_data: bool, optional
		Whether to check properties of the data and potentially recast it to
		torch.tensors. This does not prevent checking of parameters but can
		slightly speed up computation when you know that your inputs are valid.
		Setting this to False is also necessary for compiling.
	"""

	def __init__(self, probs=None, n_categories=None, pseudocount=0.0, 
		inertia=0.0, frozen=False, check_data=True):
		super().__init__(inertia=inertia, frozen=frozen, check_data=check_data)
		self.name = "Categorical"

		self.probs = _check_parameter(_cast_as_parameter(probs), "probs", 
			min_value=0, max_value=1, ndim=2)

		self.pseudocount = pseudocount

		self._initialized = probs is not None
		self.d = self.probs.shape[-2] if self._initialized else None

		if n_categories is not None:
			self.n_keys = n_categories
		else:
			self.n_keys = self.probs.shape[-1] if self._initialized else None
		
		self._reset_cache()

	def _initialize(self, d, n_keys):
		"""Initialize the probability distribution.

		This method is meant to only be called internally. It initializes the
		parameters of the distribution and stores its dimensionality. For more
		complex methods, this function will do more.


		Parameters
		----------
		d: int
			The dimensionality the distribution is being initialized to.

		n_keys: int
			The number of keys the distribution is being initialized with.
		"""

		self.probs = _cast_as_parameter(torch.zeros(d, n_keys, 
			dtype=self.dtype, device=self.device))

		self.n_keys = n_keys
		self._initialized = True
		super()._initialize(d)

	def _reset_cache(self):
		"""Reset the internally stored statistics.

		This method is meant to only be called internally. It resets the
		stored statistics used to update the model parameters as well as
		recalculates the cached values meant to speed up log probability
		calculations.
		"""

		if self._initialized == False:
			return

		self.register_buffer("_w_sum", torch.zeros(self.d, device=self.device))
		self.register_buffer("_xw_sum", torch.zeros(self.d, self.n_keys, 
			device=self.device))

		self.register_buffer("_log_probs", torch.log(self.probs))

	def sample(self, n):
		"""Sample from the probability distribution.

		This method will return `n` samples generated from the underlying
		probability distribution.


		Parameters
		----------
		n: int
			The number of samples to generate.
		

		Returns
		-------
		X: torch.tensor, shape=(n, self.d)
			Randomly generated samples.
		"""

		return torch.distributions.Categorical(self.probs).sample([n])

	def log_probability(self, X):
		"""Calculate the log probability of each example.

		This method calculates the log probability of each example given the
		parameters of the distribution. The examples must be given in a 2D
		format. For a categorical distribution, each entry in the data must
		be an integer in the range [0, n_keys).

		Note: This differs from some other log probability calculation
		functions, like those in torch.distributions, because it is not
		returning the log probability of each feature independently, but rather
		the total log probability of the entire example.


		Parameters
		----------
		X: list, tuple, numpy.ndarray, torch.Tensor, shape=(-1, self.d)
			A set of examples to evaluate.


		Returns
		-------
		logp: torch.Tensor, shape=(-1,)
			The log probability of each example.
		"""

		X = _check_parameter(_cast_as_tensor(X), "X", min_value=0.0,
			max_value=self.n_keys-1, ndim=2, shape=(-1, self.d),
			check_parameter=self.check_data)

		logps = torch.zeros(X.shape[0], dtype=self.probs.dtype, 
			device=self.device)
		
		for i in range(self.d):
			if isinstance(X, torch.masked.MaskedTensor):
				logp_ = self._log_probs[i][X[:, i]._masked_data]
				logp_[~X[:, i]._masked_mask] = 0
				logps += logp_
			else:
				logps += self._log_probs[i][X[:, i]]

		return logps

	def summarize(self, X, sample_weight=None):
		"""Extract the sufficient statistics from a batch of data.

		This method calculates the sufficient statistics from optionally
		weighted data and adds them to the stored cache. The examples must be
		given in a 2D format. Sample weights can either be provided as one
		value per example or as a 2D matrix of weights for each feature in
		each example.


		Parameters
		----------
		X: list, tuple, numpy.ndarray, torch.Tensor, shape=(-1, self.d)
			A set of examples to summarize.

		sample_weight: list, tuple, numpy.ndarray, torch.Tensor, optional
			A set of weights for the examples. This can be either of shape
			(-1, self.d) or a vector of shape (-1,). Default is ones.
		"""

		if self.frozen == True:
			return

		X = _cast_as_tensor(X)
		if not self._initialized:
			if self.n_keys is not None:
				n_keys = self.n_keys
			elif isinstance(X, torch.masked.MaskedTensor):
				n_keys = int(torch.max(X._masked_data)) + 1
			else:
				n_keys = int(torch.max(X)) + 1

			self._initialize(X.shape[1], n_keys)

		X = _check_parameter(X, "X", min_value=0, max_value=self.n_keys-1, 
			ndim=2, shape=(-1, self.d), check_parameter=self.check_data)
		sample_weight = _reshape_weights(X, _cast_as_tensor(sample_weight))

		_inplace_add(self._w_sum, torch.sum(sample_weight, dim=0))
		for i in range(self.n_keys):
			_inplace_add(self._xw_sum[:, i], torch.sum((X == i) * sample_weight, 
				dim=0))

	def from_summaries(self):
		"""Update the model parameters given the extracted statistics.

		This method uses calculated statistics from calls to the `summarize`
		method to update the distribution parameters. Hyperparameters for the
		update are passed in at initialization time.

		Note: Internally, a call to `fit` is just a successive call to the
		`summarize` method followed by the `from_summaries` method.
		"""

		if self.frozen == True:
			return

		probs = (self._xw_sum + self.pseudocount) / (self._w_sum + 
			self.pseudocount * self.n_keys).unsqueeze(1)

		_update_parameter(self.probs, probs, self.inertia)
		self._reset_cache()



# _utils.py
# Jacob Schreiber <jmschreiber91@gmail.com>

import numpy
import torch
import collections

from apricot import FacilityLocationSelection
from apricot import FeatureBasedSelection


eps = torch.finfo(torch.float32).eps

class BufferList(torch.nn.Module):
	"""A buffer list."""

	def __init__(self, buffers):
		super(BufferList, self).__init__()

		self.buffers = []
		for i, b in enumerate(buffers):
			name = "_buffer_{}".format(i)
			
			self.register_buffer(name, b)
			self.buffers.append(getattr(self, name))

	def __repr__(self):
		return str(self.buffers)

	def __getitem__(self, i):
		return self.buffers[i]

	@property
	def dtype(self):
		return self.buffers[0].dtype


def _inplace_add(X, Y):
	"""Do an in-place addition on X accounting for if Y is a masked tensor."""

	if isinstance(Y, torch.masked.MaskedTensor):
		X += Y._masked_data
	else:
		X += Y

def _cast_as_tensor(value, dtype=None):
	"""Set the parameter.""" 

	if value is None:
		return None

	if isinstance(value, (torch.nn.Parameter, torch.Tensor, torch.masked.MaskedTensor)):
		if dtype is None:
			return value
		elif value.dtype == dtype:
			return value
		else:
			return value.type(dtype)
			
	if isinstance(value, list) and all(isinstance(v, numpy.ndarray) for v in value):
		value = numpy.array(value)
		
	if isinstance(value, (float, int, list, tuple, numpy.ndarray)):
		if dtype is None:
			return torch.tensor(value)
		else:
			return torch.tensor(value, dtype=dtype)


def _cast_as_parameter(value, dtype=None, requires_grad=False):
	"""Set the parameter.""" 

	if value is None:
		return None

	value = _cast_as_tensor(value, dtype=dtype)
	return torch.nn.Parameter(value, requires_grad=requires_grad)


def _update_parameter(value, new_value, inertia=0.0, frozen=None):
	"""Update a parameter unles.
	"""

	if hasattr(value, "frozen") and getattr(value, "frozen") == True:
		return

	if inertia == 0.0:
		value[...] = _cast_as_parameter(new_value)

	elif inertia < 1.0:
		value_ = inertia*value + (1-inertia)*new_value

		inf_idx = torch.isinf(value)
		inf_idx_new = torch.isinf(new_value)

		value_[inf_idx] = value[inf_idx].type(value_.dtype)
		value_[inf_idx_new] = new_value[inf_idx_new].type(value_.dtype)
		
		value[:] = _cast_as_parameter(value_)


def _check_parameter(parameter, name, min_value=None, max_value=None, 
	value_sum=None, value_sum_dim=None, value_set=None, dtypes=None, ndim=None, 
	shape=None, check_parameter=True, epsilon=1e-6):
	"""Ensures that the parameter falls within a valid range.

	This check accepts several optional conditions that can be used to ensure
	that the value has the desired properties. If these conditions are set to
	`None`, they are not checked. 

	Note: `parameter` can be a single value or it can be a whole tensor/array
	of values. These checks are run either against the entire tensor, e.g.
	ndims, or against each of the values in the parameter.

	
	Parameters
	----------
	parameter: anything
		The parameter meant to be checked

	name: str
		The name of the parameter for error logging purposes.

	min_value: float or None, optional
		The minimum numeric value that any values in the parameter can take.
		Default is None.

	max_value: float or None, optional
		The maximum numeric value that any values in the parameter can take.
		Default is None.

	value_sum: float or None, optional
		The approximate sum, within eps, of the parameter. Default is None.

	value_sum_dim: int or float, optional
		What dimension to sum over. If None, sum over entire tensor. Default
		is None.

	value_set: tuple or list or set or None, optional
		The set of values that each element in the parameter can take. Default
		is None.

	dtypes: tuple or list or set or None, optional
		The set of dtypes that the parameter can take. Default is None.

	ndim: int or list or tuple or None, optional
		The number of dimensions of the tensor. Should not be used when the
		parameter is a single value. Default is None.

	shape: tuple or None, optional
		The shape of the parameter. -1 can be used to accept any value for that
		dimension.

	epsilon: float, optional
		When using `value_sum`, this is the maximum difference acceptable.
		Default is 1e-6.
	"""

	vector = (numpy.ndarray, torch.Tensor, torch.nn.Parameter)

	if parameter is None:
		return None

	if check_parameter == False:
		return parameter

	if dtypes is not None:
		if isinstance(parameter, vector):
			if parameter.dtype not in dtypes:
				raise ValueError("Parameter {} dtype must be one of {}".format(
					name, dtypes))
		else:
			if type(parameter) not in dtypes:
				raise ValueError("Parameter {} dtype must be one of {}".format(
					name, dtypes))


	if min_value is not None:
		if isinstance(parameter, vector):
			if (parameter < min_value).sum() > 0:
				raise ValueError("Parameter {} must have a minimum value above"
					" {}".format(name, min_value))
		else:
			if parameter < min_value:
				raise ValueError("Parameter {} must have a minimum value above"
					" {}".format(name, min_value))


	if max_value is not None:
		if isinstance(parameter, vector):
			if (parameter > max_value).sum() > 0:
				raise ValueError("Parameter {} must have a maximum value below"
					" {}".format(name, max_value))
		else:
			if parameter > max_value:
				raise ValueError("Parameter {} must have a maximum value below"
					" {}".format(name, max_value))


	if value_sum is not None:
		if isinstance(parameter, vector):
			if value_sum_dim is None:
				delta = torch.sum(parameter) - value_sum
			else:
				delta = torch.sum(parameter, dim=value_sum_dim) - value_sum

			if torch.any(torch.abs(delta) > epsilon):
				raise ValueError("Parameter {} must sum to {}".format(name, 
					value_sum))
		else:
			if abs(parameter - value_sum) > epsilon:
				raise ValueError("Parameter {} must sum to {}".format(name, 
					value_sum))


	if value_set is not None:
		if isinstance(parameter, vector):
			if (~numpy.isin(parameter, value_set)).sum() > 0:
				raise ValueError("Parameter {} must contain values in set"
					" {}".format(name, value_set))
		else:
			if parameter not in value_set:
				raise ValueError("Parameter {} must contain values in set"
					" {}".format(name, value_set))


	if ndim is not None:
		if isinstance(parameter, vector):
			if isinstance(ndim, int):
				if len(parameter.shape) != ndim:
					raise ValueError("Parameter {} must have {} dims".format(
						name, ndim))
			else:
				if len(parameter.shape) not in ndim:
					raise ValueError("Parameter {} must have {} dims".format(
						name, ndim))
		else:
			if ndim != 0:
				raise ValueError("Parameter {} must have {} dims".format(
					name, ndim))

	if shape is not None:
		if isinstance(parameter, vector):
			if len(parameter.shape) != len(shape):
				raise ValueError("Parameter {} must have shape {}".format(
					name, shape))

			for i in range(len(shape)):
				if shape[i] != -1 and shape[i] != parameter.shape[i]:
					raise ValueError("Parameter {} must have shape {}".format(
						name, shape))

	return parameter


def _check_shapes(parameters, names):
	"""Check the shapes of a set of parameters.

	This function takes in a set of parameters, as well as their names, and
	checks that the shape is correct. It will raise an error if the lengths
	of the parameters without the value of None are not equal.


	Parameters
	----------
	parameters: list or tuple
		A set of parameters, which can be None, to check the shape of.

	names: list or tuple
		A set of parameter names to refer to if something is wrong.
	"""

	n = len(parameters)

	for i in range(n):
		for j in range(n):
			if parameters[i] is None:
				continue

			if parameters[j] is None:
				continue

			n1, n2 = names[i], names[j]
			if len(parameters[i]) != len(parameters[j]):
				raise ValueError("Parameters {} and {} must be the same "
					"shape.".format(names[i], names[j]))


def _reshape_weights(X, sample_weight, device='cpu'):
	"""Handle a sample weight tensor by creating and reshaping it.

	This function will take any weight input, including None, 1D weights, and
	2D weights, and shape it into a 2D matrix with the same shape as the data
	X also passed in.

	Both elements must be PyTorch tensors.


	Parameters
	----------
	X: torch.tensor, ndims=2
		The data being weighted. The contents of this tensor are not used, only
		the shape is.

	sample_weight: torch.tensor or None
		The weight for each element in the data or None.


	Returns
	-------
	sample_weight: torch.tensor, shape=X.shape
		A tensor with the same dimensions as X with elements repeated as
		necessary.
	"""

	if sample_weight is None:
		if torch.is_floating_point(X):
			sample_weight = torch.ones(1, device=device, 
				dtype=X.dtype).expand(*X.shape)
		else:
			sample_weight = torch.ones(1, device=device, 
				dtype=torch.float32).expand(*X.shape)
	else:
		if not torch.is_floating_point(sample_weight):
			sample_weight = sample_weight.type(torch.float32)


	if len(sample_weight.shape) == 1: 
		sample_weight = sample_weight.reshape(-1, 1).expand(-1, X.shape[1])
		_check_parameter(sample_weight, "sample_weight", min_value=0)

	elif sample_weight.shape[1] == 1:
		sample_weight = sample_weight.expand(-1, X.shape[1])
		_check_parameter(sample_weight, "sample_weight", min_value=0)

	if isinstance(X, torch.masked.MaskedTensor):
		if not isinstance(sample_weight, torch.masked.MaskedTensor):
			sample_weight = torch.masked.MaskedTensor(sample_weight, 
				mask=X._masked_mask)

	_check_parameter(sample_weight, "sample_weight", shape=X.shape, 
		ndim=X.ndim)
	return sample_weight


def _initialize_centroids(X, k, algorithm='first-k', random_state=None):
	if isinstance(k, torch.Tensor):
		k = k.item()
		
	if not isinstance(random_state, numpy.random.mtrand.RandomState):
		random_state = numpy.random.RandomState(random_state)

	if algorithm == 'first-k':
		return _cast_as_tensor(torch.clone(X[:k]), dtype=torch.float32)

	elif algorithm == 'random':
		idxs = random_state.choice(len(X), size=k, replace=False)
		return _cast_as_tensor(torch.clone(X[idxs]), dtype=torch.float32)

	elif algorithm == 'submodular-facility-location':
		selector = FacilityLocationSelection(k, random_state=random_state)
		return _cast_as_tensor(selector.fit_transform(X), dtype=torch.float32)

	elif algorithm == 'submodular-feature-based':
		selector = FeatureBasedSelection(k, random_state=random_state)
		return selector.fit_transform(X)


def partition_sequences(X, sample_weight=None, priors=None, n_dists=None):
	"""Partition a set of sequences into blobs of equal length.

	This function will take in a list of sequences, where each sequence is
	a different length, and group together sequences of the same length so that
	batched operations can be more efficiently done on them. 

	Alternatively, it can take in sequences in the correct format and simply 
	return them. The correct form is to be either a single tensor that has
	three dimensions or a list of three dimensional tensors, where each
	tensor contains all the sequences of the same length.

	More concretely, the input to this function can be:

		- A single 3D tensor, or blob of data that can be cast into a 3D
		tensor, which gets returned

		- A list of 3D tensors, or elements that can be cast into 3D tensors, 
		which also gets returned

		- A list of 2D tensors, which get partitioned into a list of 3D tensors


	Parameters
	----------
	X: list or tensor
		The input sequences to be partitioned if in an incorrect format.

	sample_weight: list or tensor or None, optional
		The input sequence weights for the sequences or None. If None, return
		None. Default is None.

	priors: list or tensor or None
		The input sequence priors for the sequences or None. If None, return
		None. Default is None.

	n_dists: int or None
		The expected last dimension of the priors tensor. Must be provided if
		`priors` is provided. Default is None.


	Returns
	-------
	X_: list or tensor
		The partitioned and grouped sequences.
	"""

	if priors is not None and n_dists is None:
		raise RuntimeError("If priors are provided, n_dists must be provided as well.")
	
	# If a 3D tensor has been passed in, return it
	try:
		X = [_check_parameter(_cast_as_tensor(X), "X", ndim=3)]
	except:
		pass
	else:
		if sample_weight is not None:
			sample_weight = [_check_parameter(_cast_as_tensor(sample_weight),
				"sample_weight", min_value=0.0).to(X[0].device)]

		if priors is not None:
			priors = [_check_parameter(_cast_as_tensor(priors), "priors",
				ndim=3, shape=(*X[0].shape[:-1], n_dists)).to(X[0].device)]

		return X, sample_weight, priors

	# Otherwise, cast all elements in the list as a tensor
	X = [_cast_as_tensor(x) for x in X]

	# If a list of 3D tensors has been passed in, return it
	try:
		X = [_check_parameter(x, "X", ndim=3) for x in X]
	except:
		pass
	else:
		if sample_weight is not None:
			sample_weight = [_check_parameter(_cast_as_tensor(w_),
				"sample_weight", min_value=0.0).to(X[0].device) 
					for w_ in sample_weight]

		if priors is not None:
			priors = [_check_parameter(_cast_as_tensor(p), "priors",
				ndim=3, shape=(*X[i].shape[:-1], n_dists)).to(X[0].device) 
					for i, p in enumerate(priors)]

		if all([x.ndim == 3 for x in X]):
			return X, sample_weight, priors

	# Otherwise, group together same-sized examples
	X_dict = collections.defaultdict(list)
	sample_weight_dict = collections.defaultdict(list)
	priors_dict = collections.defaultdict(list)

	for i, x in enumerate(X):
		x = _check_parameter(x, "X", ndim=2)
		n = len(x)

		X_dict[n].append(x)

		if sample_weight is not None:
			w = _check_parameter(_cast_as_tensor(sample_weight[i]), 
				"sample_weight", min_value=0.0).to(X[0].device)

			sample_weight_dict[n].append(w)

		if priors is not None:
			p = _check_parameter(_cast_as_tensor(priors[i]), "priors",
				ndim=2, shape=(*x.shape[:-1], n_dists)).to(X[0].device)

			priors_dict[n].append(p)

	keys = sorted(X_dict.keys())
	
	X_ = [torch.stack(X_dict[key]) for key in keys]

	if sample_weight is None:
		sample_weight_ = None
	else:
		sample_weight_ = [torch.stack(sample_weight_dict[key]) for key in keys]

	if priors is None:
		priors_ = None
	else:
		priors_ = [torch.stack(priors_dict[key]) for key in keys]

	return X_, sample_weight_, priors_



# normal.py
# Contact: Jacob Schreiber <jmschreiber91@gmail.com>

import torch

from .._utils import _cast_as_tensor
from .._utils import _cast_as_parameter
from .._utils import _update_parameter
from .._utils import _check_parameter
from .._utils import _check_shapes

from ._distribution import Distribution


# Define some useful constants
NEGINF = float("-inf")
INF = float("inf")
SQRT_2_PI = 2.50662827463
LOG_2_PI = 1.83787706641


class Normal(Distribution):
	"""A normal distribution object.

	A normal distribution models the probability of a variable occurring under
	a bell-shaped curve. It is described by a vector of mean values and a
	covariance value that can be zero, one, or two dimensional. This
	distribution can assume that features are independent of the others if
	the covariance type is 'diag' or 'sphere', but if the type is 'full' then
	the features are not independent.

	There are two ways to initialize this object. The first is to pass in
	the tensor of probability parameters, at which point they can immediately be
	used. The second is to not pass in the rate parameters and then call
	either `fit` or `summary` + `from_summaries`, at which point the probability
	parameter will be learned from data.


	Parameters
	----------
	means: list, numpy.ndarray, torch.Tensor or None, shape=(d,), optional
		The mean values of the distributions. Default is None.

	covs: list, numpy.ndarray, torch.Tensor, or None, optional
		The variances and covariances of the distribution. If covariance_type
		is 'full', the shape should be (self.d, self.d); if 'diag', the shape
		should be (self.d,); if 'sphere', it should be (1,). Note that this is
		the variances or covariances in all settings, and not the standard
		deviation, as may be more common for diagonal covariance matrices.
		Default is None.

	covariance_type: str, optional
		The type of covariance matrix. Must be one of 'full', 'diag', or
		'sphere'. Default is 'full'. 

	min_cov: float or None, optional
		The minimum variance or covariance.

	inertia: float, [0, 1], optional
		Indicates the proportion of the update to apply to the parameters
		during training. When the inertia is 0.0, the update is applied in
		its entirety and the previous parameters are ignored. When the
		inertia is 1.0, the update is entirely ignored and the previous
		parameters are kept, equivalently to if the parameters were frozen.

	frozen: bool, optional
		Whether all the parameters associated with this distribution are frozen.
		If you want to freeze individual pameters, or individual values in those
		parameters, you must modify the `frozen` attribute of the tensor or
		parameter directly. Default is False.
	"""

	def __init__(self, means=None, covs=None, covariance_type='full', 
		min_cov=None, inertia=0.0, frozen=False, check_data=True):
		super().__init__(inertia=inertia, frozen=frozen, check_data=check_data)
		self.name = "Normal"

		self.means = _check_parameter(_cast_as_parameter(means), "means", 
			ndim=1)
		self.covs = _check_parameter(_cast_as_parameter(covs), "covs", 
			ndim=(1, 2))

		_check_shapes([self.means, self.covs], ["means", "covs"])

		self.min_cov = _check_parameter(min_cov, "min_cov", min_value=0, ndim=0)
		self.covariance_type = covariance_type

		self._initialized = (means is not None) and (covs is not None)
		self.d = self.means.shape[-1] if self._initialized else None
		self._reset_cache()

	def _initialize(self, d):
		"""Initialize the probability distribution.

		This method is meant to only be called internally. It initializes the
		parameters of the distribution and stores its dimensionality. For more
		complex methods, this function will do more.


		Parameters
		----------
		d: int
			The dimensionality the distribution is being initialized to.
		"""

		self.means = _cast_as_parameter(torch.zeros(d, dtype=self.dtype,
			device=self.device))
		
		if self.covariance_type == 'full':
			self.covs = _cast_as_parameter(torch.zeros(d, d, 
				dtype=self.dtype, device=self.device))
		elif self.covariance_type == 'diag':
			self.covs = _cast_as_parameter(torch.zeros(d, dtype=self.dtype,
				device=self.device))
		elif self.covariance_type == 'sphere':
			self.covs = _cast_as_parameter(torch.tensor(0, dtype=self.dtype,
				device=self.device))

		self._initialized = True
		super()._initialize(d)

	def _reset_cache(self):
		"""Reset the internally stored statistics.

		This method is meant to only be called internally. It resets the
		stored statistics used to update the model parameters as well as
		recalculates the cached values meant to speed up log probability
		calculations.
		"""

		if self._initialized == False:
			return

		self.register_buffer("_w_sum", torch.zeros(self.d, dtype=self.dtype, 
			device=self.device))
		self.register_buffer("_xw_sum", torch.zeros(self.d, dtype=self.dtype,
			device=self.device))

		if self.covariance_type == 'full':
			self.register_buffer("_xxw_sum", torch.zeros(self.d, self.d, 
				dtype=self.dtype, device=self.device))

			if self.covs.sum() > 0.0:
				chol = torch.linalg.cholesky(self.covs)
				_inv_cov = torch.linalg.solve_triangular(chol, torch.eye(
					len(self.covs), dtype=self.dtype, device=self.device), 
					upper=False).T
				_inv_cov_dot_mu = torch.matmul(self.means, _inv_cov)
				_log_det = -0.5 * torch.linalg.slogdet(self.covs)[1]
				_theta = _log_det - 0.5 * (self.d * LOG_2_PI)

				self.register_buffer("_inv_cov", _inv_cov)
				self.register_buffer("_inv_cov_dot_mu", _inv_cov_dot_mu)
				self.register_buffer("_log_det", _log_det)
				self.register_buffer("_theta", _theta)

		elif self.covariance_type in ('diag', 'sphere'):
			self.register_buffer("_xxw_sum", torch.zeros(self.d, 
				dtype=self.dtype, device=self.device))

			if self.covs.sum() > 0.0:
				_log_sigma_sqrt_2pi = -torch.log(torch.sqrt(self.covs) * 
					SQRT_2_PI)
				_inv_two_sigma = 1. / (2 * self.covs)

				self.register_buffer("_log_sigma_sqrt_2pi", _log_sigma_sqrt_2pi)
				self.register_buffer("_inv_two_sigma", _inv_two_sigma)

			if torch.any(self.covs < 0):
				raise ValueError("Variances must be positive.")

	def sample(self, n):
		"""Sample from the probability distribution.

		This method will return `n` samples generated from the underlying
		probability distribution.


		Parameters
		----------
		n: int
			The number of samples to generate.
		

		Returns
		-------
		X: torch.tensor, shape=(n, self.d)
			Randomly generated samples.
		"""

		if self.covariance_type == 'diag':
			return torch.distributions.Normal(self.means, self.covs).sample([n])
		elif self.covariance_type == 'full':
			return torch.distributions.MultivariateNormal(self.means, 
				self.covs).sample([n])

	def log_probability(self, X):
		"""Calculate the log probability of each example.

		This method calculates the log probability of each example given the
		parameters of the distribution. The examples must be given in a 2D
		format.

		Note: This differs from some other log probability calculation
		functions, like those in torch.distributions, because it is not
		returning the log probability of each feature independently, but rather
		the total log probability of the entire example.


		Parameters
		----------
		X: list, tuple, numpy.ndarray, torch.Tensor, shape=(-1, self.d)
			A set of examples to evaluate.


		Returns
		-------
		logp: torch.Tensor, shape=(-1,)
			The log probability of each example.
		"""

		X = _check_parameter(_cast_as_tensor(X, dtype=self.means.dtype), "X", 
			ndim=2, shape=(-1, self.d), check_parameter=self.check_data)

		if self.covariance_type == 'full':
			logp = torch.matmul(X, self._inv_cov) - self._inv_cov_dot_mu
			logp = self.d * LOG_2_PI + torch.sum(logp ** 2, dim=-1)
			logp = self._log_det - 0.5 * logp
			return logp
		
		elif self.covariance_type in ('diag', 'sphere'):
			return torch.sum(self._log_sigma_sqrt_2pi - ((X - self.means) ** 2) 
				* self._inv_two_sigma, dim=-1)

	def summarize(self, X, sample_weight=None):
		"""Extract the sufficient statistics from a batch of data.

		This method calculates the sufficient statistics from optionally
		weighted data and adds them to the stored cache. The examples must be
		given in a 2D format. Sample weights can either be provided as one
		value per example or as a 2D matrix of weights for each feature in
		each example.


		Parameters
		----------
		X: list, tuple, numpy.ndarray, torch.Tensor, shape=(-1, self.d)
			A set of examples to summarize.

		sample_weight: list, tuple, numpy.ndarray, torch.Tensor, optional
			A set of weights for the examples. This can be either of shape
			(-1, self.d) or a vector of shape (-1,). Default is ones.
		"""

		if self.frozen == True:
			return

		X, sample_weight = super().summarize(X, sample_weight=sample_weight)
		X = _cast_as_tensor(X, dtype=self.means.dtype)
		sample_weight = _cast_as_tensor(sample_weight, dtype=self.means.dtype)
		if self.covariance_type == 'full':
			self._w_sum += torch.sum(sample_weight, dim=0)
			self._xw_sum += torch.sum(X * sample_weight, axis=0)
			self._xxw_sum += torch.matmul((X * sample_weight).T, X)

		elif self.covariance_type in ('diag', 'sphere'):
			self._w_sum[:] = self._w_sum + torch.sum(sample_weight, dim=0)
			self._xw_sum[:] = self._xw_sum + torch.sum(X * sample_weight, dim=0)
			self._xxw_sum[:] = self._xxw_sum + torch.sum(X ** 2 * 
				sample_weight, dim=0)

	def from_summaries(self):
		"""Update the model parameters given the extracted statistics.

		This method uses calculated statistics from calls to the `summarize`
		method to update the distribution parameters. Hyperparameters for the
		update are passed in at initialization time.

		Note: Internally, a call to `fit` is just a successive call to the
		`summarize` method followed by the `from_summaries` method.
		"""

		if self.frozen == True:
			return

		means = self._xw_sum / self._w_sum

		if self.covariance_type == 'full':
			v = self._xw_sum.unsqueeze(0) * self._xw_sum.unsqueeze(1)
			covs = self._xxw_sum / self._w_sum -  v / self._w_sum ** 2.0

		elif self.covariance_type in ['diag', 'sphere']:
			covs = self._xxw_sum / self._w_sum - \
				self._xw_sum ** 2.0 / self._w_sum ** 2.0
			if self.covariance_type == 'sphere':
				covs = covs.mean(dim=-1)

		_update_parameter(self.means, means, self.inertia)
		_update_parameter(self.covs, covs, self.inertia)
		self._reset_cache()



from setuptools import setup

setup(
	name='pomegranate',
	version='1.1.2',
	author='Jacob Schreiber',
	author_email='jmschreiber91@gmail.com',
	packages=['pomegranate', 'pomegranate.distributions', 'pomegranate.hmm'],
	url='https://github.com/jmschrei/torchegranate',
	license='MIT',
	description='A PyTorch implementation of probabilistic models.',
	install_requires=[
		'numpy >= 1.22.2', 
		'scipy >= 1.6.2',
		'scikit-learn >= 1.0.2',
		'torch >= 1.9.0',
		'apricot-select >= 0.6.1',
		'networkx >= 2.8.4'
	]
)


# normal.py
# Contact: Jacob Schreiber <jmschreiber91@gmail.com>

import torch

from .._utils import _cast_as_tensor
from .._utils import _cast_as_parameter
from .._utils import _update_parameter
from .._utils import _check_parameter
from .._utils import _check_shapes

from ._distribution import Distribution
from .normal import Normal


# Define some useful constants
LOG_2 = 0.6931471805599453


class HalfNormal(Normal):
    """A half-normal distribution object.

    A half-normal distribution is a distribution over positive real numbers that
    is zero for negative numbers. It is defined by a single parameter, sigma,
    which is the standard deviation of the distribution. The mean of the
    distribution is sqrt(2/pi) * sigma, and the variance is (1 - 2/pi) * sigma^2.

    This distribution can assume that features are independent of the others if
    the covariance type is 'diag' or 'sphere', but if the type is 'full' then
    the features are not independent.

    There are two ways to initialize this object. The first is to pass in
    the tensor of probablity parameters, at which point they can immediately be
    used. The second is to not pass in the rate parameters and then call
    either `fit` or `summarize` + `from_summaries`, at which point the probability
    parameter will be learned from data.


    Parameters
    ----------
    covs: list, numpy.ndarray, torch.Tensor, or None, optional
            The variances and covariances of the distribution. If covariance_type
            is 'full', the shape should be (self.d, self.d); if 'diag', the shape
            should be (self.d,); if 'sphere', it should be (1,). Note that this is
            the variances or covariances in all settings, and not the standard
            deviation, as may be more common for diagonal covariance matrices.
            Default is None.

    covariance_type: str, optional
            The type of covariance matrix. Must be one of 'full', 'diag', or
            'sphere'. Default is 'full'.

    min_cov: float or None, optional
            The minimum variance or covariance.

    inertia: float, [0, 1], optional
            Indicates the proportion of the update to apply to the parameters
            during training. When the inertia is 0.0, the update is applied in
            its entirety and the previous parameters are ignored. When the
            inertia is 1.0, the update is entirely ignored and the previous
            parameters are kept, equivalently to if the parameters were frozen.

    frozen: bool, optional
            Whether all the parameters associated with this distribution are frozen.
            If you want to freeze individual pameters, or individual values in those
            parameters, you must modify the `frozen` attribute of the tensor or
            parameter directly. Default is False.
    """

    def __init__(
        self,
        covs=None,
        covariance_type="full",
        min_cov=None,
        inertia=0.0,
        frozen=False,
        check_data=True,
    ):
        self.name = "HalfNormal"
        super().__init__(
            means=None,
            covs=covs,
            min_cov=min_cov,
            covariance_type=covariance_type,
            inertia=inertia,
            frozen=frozen,
            check_data=check_data,
        )

    def _initialize(self, d):
        """Initialize the probability distribution.

        This method is meant to only be called internally. It initializes the
        parameters of the distribution and stores its dimensionality. For more
        complex methods, this function will do more.


        Parameters
        ----------
        d: int
                The dimensionality the distribution is being initialized to.
        """
        super()._initialize(d)

    def _reset_cache(self):
        """Reset the internally stored statistics.

        This method is meant to only be called internally. It resets the
        stored statistics used to update the model parameters as well as
        recalculates the cached values meant to speed up log probability
        calculations.
        """
        super()._reset_cache()

    def sample(self, n):
        """Sample from the probability distribution.

        This method will return `n` samples generated from the underlying
        probability distribution.


        Parameters
        ----------
        n: int
                The number of samples to generate.


        Returns
        -------
        X: torch.tensor, shape=(n, self.d)
                Randomly generated samples.
        """
        if self.covariance_type in ["diag", "full"]:
            return torch.distributions.HalfNormal(self.covs).sample([n])

    def log_probability(self, X):
        """Calculate the log probability of each example.

        This method calculates the log probability of each example given the
        parameters of the distribution. The examples must be given in a 2D
        format.

        Note: This differs from some other log probability calculation
        functions, like those in torch.distributions, because it is not
        returning the log probability of each feature independently, but rather
        the total log probability of the entire example.


        Parameters
        ----------
        X: list, tuple, numpy.ndarray, torch.Tensor, shape=(-1, self.d)
                A set of examples to evaluate.


        Returns
        -------
        logp: torch.Tensor, shape=(-1,)
                The log probability of each example.
        """

        X = _check_parameter(
            _cast_as_tensor(X, dtype=self.covs.dtype),
            "X",
            ndim=2,
            shape=(-1, self.d),
            check_parameter=self.check_data,
        )
        return super().log_probability(X) + LOG_2

    def summarize(self, X, sample_weight=None):
        """Extract the sufficient statistics from a batch of data.

        This method calculates the sufficient statistics from optionally
        weighted data and adds them to the stored cache. The examples must be
        given in a 2D format. Sample weights can either be provided as one
        value per example or as a 2D matrix of weights for each feature in
        each example.


        Parameters
        ----------
        X: list, tuple, numpy.ndarray, torch.Tensor, shape=(-1, self.d)
                A set of examples to summarize.

        sample_weight: list, tuple, numpy.ndarray, torch.Tensor, optional
                A set of weights for the examples. This can be either of shape
                (-1, self.d) or a vector of shape (-1,). Default is ones.
        """

        super().summarize(X, sample_weight=sample_weight)

    def from_summaries(self):
        """Update the model parameters given the extracted statistics.

        This method uses calculated statistics from calls to the `summarize`
        method to update the distribution parameters. Hyperparameters for the
        update are passed in at initialization time.

        Note: Internally, a call to `fit` is just a successive call to the
        `summarize` method followed by the `from_summaries` method.
        """

        if self.frozen == True:
            return

        #  the means are always zero for a half normal distribution
        means = torch.zeros(self.d, dtype=self.covs.dtype)

        if self.covariance_type == "full":
            v = self._xw_sum.unsqueeze(0) * self._xw_sum.unsqueeze(1)
            covs = self._xxw_sum / self._w_sum - v / self._w_sum**2.0

        elif self.covariance_type in ["diag", "sphere"]:
            covs = self._xxw_sum / self._w_sum - self._xw_sum**2.0 / self._w_sum**2.0
            if self.covariance_type == "sphere":
                covs = covs.mean(dim=-1)

        _update_parameter(self.covs, covs, self.inertia)
        _update_parameter(self.means, means, self.inertia)
        self._reset_cache()
